"""Download files from dropbox."""
import logging
import os
import sys
from urllib import request
import zipfile
from warnings import warn

from .rw_config import load_config_json
from .path import get_data_url_path


logger = logging.getLogger('visbrain')


__all__ = ["download_file"]


def get_data_url(name, astype):
    """Get filename and url to a file.

    Parameters
    ----------
    name : string
        Name of the file.
    astype : string
        Type of the file to download.

    Returns
    -------
    url : string
        Url to the file to download.
    """
    # Get path to data_url.txt :
    urls = load_config_json(get_data_url_path())[astype]
    # Try to get the file :
    try:
        url_to_download = urls[name]
        if not bool(url_to_download.find('dl=1') + 1):
            warn("The dropbox url should contains 'dl=1'.")
        return urls[name]
    except:
        raise IOError(name + " not in the default path list of files.")


def reporthook(blocknum, blocksize, totalsize):
    """Report downloading status."""
    readsofar = blocknum * blocksize
    if totalsize > 0:
        percent = min(100, readsofar * 1e2 / totalsize)
        s = "\rSTATUS : %5.1f%% %*d / %d" % (
            percent, len(str(totalsize)), readsofar, totalsize)
        sys.stderr.write(s)
        if readsofar >= totalsize:  # near the end
            sys.stderr.write("\n")
    else:  # total size is unknown
        sys.stderr.write("\rread %d" % (readsofar,))


def download_file(name, astype=None, filename=None, to_path=None, unzip=False,
                  remove_archive=False, use_pwd=False):
    """Download a file.

    By default this function download a file to ~/visbrain_data.

    Parameters
    ----------
    name : string
        Name of the file to download or url.
    astype : str | None
        If name is a name of a file that can be downloaded, astype refer to the
        type of the file.
    filename : string | None
        Name of the file to be saved in case of url.
    to_path : string | None
        Download file to the path specified.
    unzip : bool | False
        Unzip archive if needed.
    remove_archive : bool | False
        Remove archive after unzip.
    use_pwd : bool | False
        Download the file to the current directory.

    Returns
    -------
    path_to_file : string
        Path to the downloaded file.
    """
    # Default visbrain-path to data (HOME/visbrain_data):
    vb_path = os.path.join(os.path.expanduser('~'), 'visbrain_data')
    vb_path = os.getcwd() if use_pwd else vb_path
    if bool(name.find('http') + 1):
        if filename is None:
            filename = os.path.split(name)[1]
        assert isinstance(filename, str)
        url = name
    else:
        assert isinstance(name, str) and isinstance(astype, str)
        filename, url = name, get_data_url(name, astype)
        to_path = os.path.join(vb_path, astype)
    to_path = vb_path if not isinstance(to_path, str) else to_path
    path_to_file = os.path.join(to_path, filename)
    to_download = not os.path.isfile(path_to_file)

    # Dowload file if needed :
    if to_download:
        logger.info('Downloading %s' % path_to_file)
        # Check if directory exists else creates it
        if not os.path.exists(to_path):
            logger.info('Folder %s created' % to_path)
            os.makedirs(to_path)
        # Download file :
        fh, _ = request.urlretrieve(url, path_to_file, reporthook=reporthook)
        # Unzip file :
        if unzip:
            # Unzip archive :
            zip_file_object = zipfile.ZipFile(fh, 'r')
            zip_file_object.extractall(path=to_path)
            zip_file_object.close()
            logger.info('Unzip archive')
            if remove_archive:  # Remove archive :
                logger.info('Archive %s removed' % path_to_file)
                os.remove(path_to_file)
    else:
        logger.info("File already dowloaded (%s)." % path_to_file)
    return path_to_file



"""Main class for info managment."""
from os import path

from PyQt5 import QtWidgets

from sleep_scoring.utils import sleepstats


class UiInfo(object):
    """Main class for info managment."""

    def __init__(self):
        """Init."""
        pass

    def _fcn_info_update(self):
        """Complete the table sleep info."""
        table = self._infoTable
        # Get sleep stats :
        stats = sleepstats(self._hyp.gui_to_hyp(), self._sf)

        # Add global informations to stats dict
        is_file = isinstance(self._file, str)
        stats['Filename'] = path.basename(self._file) if is_file else ''
        stats['Sampling frequency'] = str(self._sfori) + " Hz"
        stats['Down-sampling'] = str(self._sf) + " Hz"

        self._keysInfo = [''] * len(stats)
        self._valInfo = [''] * len(stats)
        # Check line number:
        table.setRowCount(len(stats))
        # Fill table :
        for num, (k, v) in enumerate(stats.items()):
            # Add keys :
            table.setItem(int(num), 0, QtWidgets.QTableWidgetItem(k))
            # Add values :
            table.setItem(int(num), 1, QtWidgets.QTableWidgetItem(str(v)))
            # Remember variables :
            self._keysInfo[int(num)] = k
            self._valInfo[int(num)] = str(v)



"""This script initialize all ui files."""

from .ui_settings import UiSettings
from .ui_panels import UiPanels
from .ui_info import UiInfo
from .ui_tools import UiTools
from .ui_scoring import UiScoring
from .ui_detection import UiDetection
from .ui_menu import UiMenu
from .ui_annotate import UiAnnotate
from .ui_screenshot import UiScreenshot
from sleep_scoring.config import PROFILER


class UiElements(UiSettings, UiPanels, UiInfo, UiTools, UiScoring,
                 UiDetection, UiAnnotate, UiMenu, UiScreenshot):
    """Inherit from the diffrent Ui files and initialize them."""

    def __init__(self):
        """Init."""
        for k in ['UiSettings', 'UiPanels', 'UiInfo', 'UiTools', 'UiScoring',
                  'UiDetection', 'UiAnnotate', 'UiMenu', 'UiScreenshot']:
            eval(k + '.__init__(self)')
            PROFILER("%s" % k, level=1)



"""Utility functions for MNE."""
import datetime
import numpy as np
from ..utils import get_dsf

__all__ = ['mne_switch']


def mne_switch(file, ext, downsample, preload=True, **kwargs):
    """Read sleep datasets using mne.io.

    Parameters
    ----------
    file : string
        Filename (without extension).
    ext : string
        File extension (e.g. '.edf'').
    preload : bool | True
        Preload data in memory.
    kwargs : dict | {}
        Further arguments to pass to the mne.io.read function.

    Returns
    -------
    sf : float
        The original sampling-frequency.
    downsample : float
        The down-sampling frequency used.
    dsf : int
        The down-sampling factor.
    data : array_like
        The raw data of shape (n_channels, n_points)
    channels : list
        List of channel names.
    n : int
        Number of time points before down-sampling.
    start_time : datetime.time
        The time offset.
    """
    from mne import io

    # Get full path :
    path = file + ext

    # Preload :
    if preload is False:
        preload = 'temp.dat'
    kwargs['preload'] = preload

    if ext.lower() in ['.edf', '.bdf', '.gdf']:  # EDF / BDF / GDF
        raw = io.read_raw_edf(path, **kwargs)
    elif ext.lower == '.set':   # EEGLAB
        raw = io.read_raw_eeglab(path, **kwargs)
    elif ext.lower() in ['.egi', '.mff']:  # EGI / MFF
        raw = io.read_raw_egi(path, **kwargs)
    elif ext.lower() == '.cnt':  # CNT
        raw = io.read_raw_cnt(path, **kwargs)
    elif ext.lower() == '.vhdr':  # BrainVision
        raw = io.read_raw_brainvision(path, **kwargs)
    else:
        raise IOError("File not supported by mne-python.")

    raw.pick_types(meg=True, eeg=True, ecg=True, emg=True)  # Remove stim lines
    sf = raw.info['sfreq']
    dsf, downsample = get_dsf(downsample, sf)
    channels = raw.info['ch_names']
    data = raw._data

    # Conversion Volt (MNE) to microVolt (Visbrain):
    if raw._raw_extras[0] is not None and 'units' in raw._raw_extras[0]:
        units = raw._raw_extras[0]['units'][0:data.shape[0]]
        data /= np.array(units).reshape(-1, 1)

    n = data.shape[1]
    start_time = datetime.time(0, 0, 0)  # raw.info['meas_date']
    anot = raw.annotations

    return sf, downsample, dsf, data[:, ::dsf], channels, n, start_time, anot



"""All visbrain modules based on PyQt5 should inherit from _PyQtModule."""
import os
import sip

from PyQt5 import QtGui
import logging

from .utils import set_widget_size, set_log_level
from .config import PROFILER, CONFIG
from .io import (path_to_tmp, download_file, clean_tmp, path_to_visbrain_data)

sip.setdestroyonexit(False)
logger = logging.getLogger('visbrain')


class _PyQtModule(object):
    """Shared methods across PyQt based Visbrain modules.

    Parameters
    ----------
    verbose : logging.level | None
        Verbosity level.
    to_describe : string | None
        The parent node to describe. Only available if verbose level is on
        debug.
    icon : str | None
        Name of the icon to use.
    """

    def __init__(self, verbose=None, to_describe=None, icon=None,
                 show_settings=False):
        """Init."""
        # Log level and profiler creation (if verbose='debug')
        set_log_level(verbose)
        path_to_visbrain_data()
        self._create_tmp_folder()
        if logger.level == 10:
            import faulthandler
            faulthandler.enable()
            logger.debug("Faulthandler enabled")
        self._need_description = to_describe
        if isinstance(self._need_description, str):
            self._need_description = [self._need_description]
        self._module_icon = icon
        self._show_settings = show_settings

    ###########################################################################
    #                               TMP FOLDER
    ###########################################################################

    def _path_to_tmp_folder(self):
        return path_to_tmp()

    def _create_tmp_folder(self):
        tmp_path = path_to_tmp()
        logger.debug("tmp folder created (%s)." % tmp_path)

    def _clean_tmp_folder(self):
        tmp_path = clean_tmp()
        logger.debug("tmp folder cleaned (%s)." % tmp_path)

    ###########################################################################
    #                            SHOW // CLOSE
    ###########################################################################
    def _pyqt_title(self, title, msg, symbol='='):
        """Define a PyQt title."""
        assert isinstance(title, str) and isinstance(symbol, str)
        n_symbol = 60
        start_title = int((n_symbol / 2) - (len(title) / 2)) - 1
        upper_bar = symbol * n_symbol
        title_bar = ' ' * start_title + title
        title_template = "\n%s\n%s\n\n{msg}" % (upper_bar, title_bar)
        logger.profiler(title_template.format(msg=msg))

    def show(self):
        """Display the graphical user interface."""
        # Fixed size for the settings panel :
        if hasattr(self, 'q_widget'):
            set_widget_size(CONFIG['PYQT_APP'], self.q_widget, 23)
            self.q_widget.setVisible(self._show_settings)
        # Force the quick settings tab to be on the first tab :
        if hasattr(self, 'QuickSettings'):
            self.QuickSettings.setCurrentIndex(0)
        # Set icon (if possible) :
        if isinstance(self._module_icon, str):
            path_ico = path_to_visbrain_data(self._module_icon, 'icons')
            if not os.path.isfile(path_ico):
                download_file(self._module_icon, astype='icons')
            if os.path.isfile(path_ico):
                app_icon = QtGui.QIcon()
                app_icon.addFile(path_ico)
                self.setWindowIcon(app_icon)
            else:  # don't crash just for an icon...
                logger.debug("No icon found (%s)" % self._module_icon)
        else:
            logger.debug("No icon passed as an input.")
        # Tree description if log level is on debug :
        if isinstance(self._need_description, list):
            for k in self._need_description:
                self._pyqt_title('Tree', eval('self.%s.describe_tree()' % k))
        # Show and maximized the window :
        if PROFILER and logger.level == 1:
            self._pyqt_title('Profiler', '')
            PROFILER.finish()
        # If PyQt GUI :
        if CONFIG['SHOW_PYQT_APP']:
            self.showMaximized()
            CONFIG['VISPY_APP'].run()
        # Finally clean the tmp folder :
        self._clean_tmp_folder()

    def closeEvent(self, event):  # noqa
        """Executed method when the GUI closed."""
        CONFIG['PYQT_APP'].quit()
        logger.debug("App closed.")



"""Write data.

- write_npy : Write data as a NumPy (npy) file
- write_npz : Write data as a NumPy (npz) file
- write_mat : Write data as a Matlab (mat) file
- write_txt : Write data as a text (txt) file
- write_csv : Write data as a CSV (csv) file
- write_hyp : Write data as a hyp (hyp) file
"""

__all__ = ['write_npy', 'write_npz', 'write_mat', 'write_txt', 'write_csv',
           'write_hyp']


def write_npy():
    """Write data as a NumPy (npy) file."""
    pass


def write_npz():
    """Write data as a NumPy (npz) file."""
    pass


def write_mat():
    """Write data as a Matlab (mat) file."""
    pass


def write_txt(file, data, delimiter=', '):
    """Write data as a text (txt) file.

    Parameters
    ----------
    file : string
        File name for saving file.
    data : list
        List of data to save to the txt file.
    """
    # Open file :
    ofile = open(file, 'w')
    for k in data:
        ofile.write("%s\n" % delimiter.join(k))
    return


def write_csv(file, data, delimiter=','):
    """Write data as a CSV (csv) file.

    Parameters
    ----------
    file : string
        File name for saving file.
    data : list
        List of data to save to the csv file.
    """
    import csv
    with open(file, 'w') as csvfile:
        writer = csv.writer(csvfile, dialect='excel', delimiter=delimiter)
        for k in data:
            writer.writerow(k)
    return


def write_hyp():
    """Write data as a hyp (hyp) file."""
    pass



"""Set of functions for path definitions."""
import logging
import os
import sys

logger = logging.getLogger('visbrain')


__all__ = ['path_to_visbrain_data', 'get_files_in_folders', 'path_to_tmp',
           'clean_tmp', 'get_data_url_path']


def path_to_visbrain_data(file=None, folder=None):
    """Get the path to the visbrain_data folder.

    Parameters
    ----------
    folder : string | None
        Folder name.
    file : string | None
        File name. If None, only the path to the visbrain_data folder is
        returned.

    Returns
    -------
    path : string
        Path to the file or to the visbrain_data.
    """
    vb_path = os.path.join(os.path.expanduser('~'), 'visbrain_data')
    folder = '' if not isinstance(folder, str) else folder
    vb_path = os.path.join(vb_path, folder)
    if not os.path.exists(vb_path):
        os.makedirs(vb_path)
        logger.info("visbrain_data has been added to %s" % vb_path)
    file = '' if not isinstance(file, str) else file
    return os.path.join(vb_path, file)


def get_data_url_path():
    """Get the path to the data_url JSON file."""
    url_path = sys.modules[__name__].__file__.split('io')[0]
    return os.path.join(url_path, 'data_url.json')


def get_files_in_folders(*args, with_ext=False, with_path=False, file=None,
                         exclude=None, sort=True, unique=True):
    """Get all files in several folders.

    Parameters
    ----------
    args : string
        Path to folders.
    with_ext : bool | False
        Specify if returned files should contains extensions.
    with_path : bool | False
        Specify if returned files should contains full path to it.
    file : string | None
        Specify if a specific file name is needed.
    exclude : list | None
        List of patterns to exclude
    sort : bool | True
        Sort the resulting list of files.
    unique : bool | True
        Get a unique list of files.

    Returns
    -------
    files : list
        List of files in selected folders if no file is provided. If file is a
        string, return the path to it, None if the file doesn't exist.
    """
    # Search the file :
    files = []
    if isinstance(file, str):
        import glob
        for k in args:
            if os.path.exists(k):
                files += glob.glob(os.path.join(k, file))
        return files
    # Get the list of files :
    for k in args:
        if os.path.exists(k):
            if with_path:
                files += [os.path.join(k, i) for i in os.listdir(k)]
            else:
                files += os.listdir(k)
    # Keep only a selected file :
    if isinstance(file, str) and (file in files):
        files = [files[files.index(file)]]
    # Return either files with full path or only file name :
    if not with_ext:
        files = [os.path.splitext(k)[0] for k in files]
    # Patterns to exclude :
    if isinstance(exclude, (list, tuple)):
        from itertools import product
        files = [k for k, i in product(files, exclude) if i not in k]
    # Unique :
    if unique:
        files = list(set(files))
    # Sort list :
    if sort:
        files.sort()
    return files


def path_to_tmp(file=None, folder=None):
    """Get the path to the tmp folder."""
    tmp_path = os.path.join(path_to_visbrain_data(), 'tmp')
    if not os.path.exists(tmp_path):
        os.mkdir(tmp_path)
    folder = '' if not isinstance(folder, str) else folder
    file = '' if not isinstance(file, str) else file
    tmp_path = os.path.join(tmp_path, folder)
    if not os.path.exists(tmp_path):
        os.mkdir(tmp_path)
    return os.path.join(tmp_path, file)


def clean_tmp():
    """Clean the tmp folder."""
    tmp_path = os.path.join(path_to_visbrain_data(), 'tmp')
    if os.path.exists(tmp_path):
        import shutil
        shutil.rmtree(tmp_path)



"""Load data files.

This file contain functions to load :
- Matlab (*.mat)
- Pickle (*.pickle)
- NumPy (*.npy and *.npz)
- Text (*.txt)
- CSV (*.csv)
- JSON (*.json)
"""
import numpy as np


__all__ = ('read_mat', 'read_pickle', 'read_npy', 'read_npz', 'read_txt',
           'read_csv', 'read_json', 'read_stc')


def read_mat(path, vars=None):
    """Read data from a Matlab (mat) file."""
    from scipy.io import loadmat
    return loadmat(path, variable_names=vars)


def read_pickle(path, vars=None):
    """Read data from a Pickle (pickle) file."""
    # np.loads? ou depuis import pickle
    pass


def read_npy(path):
    """Read data from a NumPy (npy) file."""
    return np.load(path)


def read_npz(path, vars=None):
    """Read data from a Numpy (npz) file."""
    pass


def read_txt(path):
    """Read data from a text (txt) file."""
    pass


def read_csv(path):
    """Read data from a CSV (csv) file."""
    pass


def read_json(path):
    """Read data from a JSON (json) file."""
    pass


def read_stc(path):
    """Read an STC file from the MNE package.

    STC files contain activations or source reconstructions
    obtained from EEG and MEG data.

    This function is a copy from the PySurfer package. See :
    https://github.com/nipy/PySurfer/blob/master/surfer/io.py

    Parameters
    ----------
    path : string
        Path to STC file

    Returns
    -------
    data : dict
        The STC structure. It has the following keys:
           tmin           The first time point of the data in seconds
           tstep          Time between frames in seconds
           vertices       vertex indices (0 based)
           data           The data matrix (nvert * ntime)
    """
    fid = open(path, 'rb')

    stc = dict()

    fid.seek(0, 2)  # go to end of file
    file_length = fid.tell()
    fid.seek(0, 0)  # go to beginning of file

    # read tmin in ms
    stc['tmin'] = float(np.fromfile(fid, dtype=">f4", count=1))
    stc['tmin'] /= 1000.0

    # read sampling rate in ms
    stc['tstep'] = float(np.fromfile(fid, dtype=">f4", count=1))
    stc['tstep'] /= 1000.0

    # read number of vertices/sources
    vertices_n = int(np.fromfile(fid, dtype=">u4", count=1))

    # read the source vector
    stc['vertices'] = np.fromfile(fid, dtype=">u4", count=vertices_n)

    # read the number of timepts
    data_n = int(np.fromfile(fid, dtype=">u4", count=1))

    if ((file_length / 4 - 4 - vertices_n) % (data_n * vertices_n)) != 0:
        raise ValueError('incorrect stc file size')

    # read the data matrix
    stc['data'] = np.fromfile(fid, dtype=">f4", count=vertices_n * data_n)
    stc['data'] = stc['data'].reshape([data_n, vertices_n]).T

    # close the file
    fid.close()
    return stc



"""Base class for objects of type source."""
import logging
import numpy as np

from vispy import scene
from vispy.scene.visuals import Arrow
from vispy.visuals.line.arrow import ARROW_TYPES

from .visbrain_obj import VisbrainObject, CombineObjects
from ..utils import array2colormap, color2vb, wrap_properties, normalize


logger = logging.getLogger('visbrain')
ARROW_DTYPES = [np.dtype([('start', float, 3), ('end', float, 3)]),
                np.dtype([('vertices', float, 3), ('normals', float, 3)])]


class VectorObj(VisbrainObject):
    """Create a vector object.

    Parameters
    ----------
    name : string
        Name of the vector object.
    arrows : array_like, tuple, list
        The position of arrows. Use either :

            * A list (or tuple) of two arrays with identical shapes (N, 3).
              The first array specify the (x, y, z) position where arrows start
              and the second the (x, y, z) position of the end of each arrow.
            * Alternatively to the point above, an array of type
              [('start', float, 3), ('end', float, 3)] can also be used.
            * An array of type [('vertices', float, 3), ('normals', float, 3)].
              This method use the normals to vertices to inferred the arrow
              locations. In addition, if `data` is not None, `data` is used to
              inferred the arrow length.
    data : array_like | None
        Attach some data to each vector. This data can be used to inferred the
        color.
    inferred_data : bool | False
        If the `arrows` input use the (start, end) method and if inferred_data
        is set to True, the magnitude of each vector is used as data.
    select : array_like | None
        An array of boolean values to select some specifics arrows.
    color : array_like/tuple/string | 'black'
        If no data are provided, use this parameter to set a unique color for
        all vectors.
    dynamic : tuple | None
        Use a dynamic transparency method. The dynamic input must be a tuple
        of two float between [0, 1]. Vectors with stronger associated data are
        going to be set more opaque.
    line_width : float | 5.
        Line width of each vector.
    arrow_size : float | 10.
        Size of the arrow-head.
    arrow_type : string | 'stealth'
        The arrow-head type. Use either 'stealth', 'curved', 'angle_30',
        'angle_60', 'angle_90', 'triangle_30', 'triangle_60', 'triangle_90'
        or 'inhibitor_round'.
    arrow_coef : float | 1.
        Use this coefficient to define longer arrows. Must be a float superior
        to 1.
    arrow_norm : tuple | (5., 20.)
        Control arrow length for arrows defined using vertices and normals.
    antialias : bool | False
        Use smoothed lines.
    cmap : string | 'viridis'
        The colormap to use (if data is not None).
    clim : tuple | None
        Colorbar limits. If None, the (max, min) of data is used (if data is
        not None).
    vmin : float | None
        Minimum threshold (if data is not None).
    under : string | 'gray'
        Color for values under vmin (if data is not None).
    vmax : float | None
        Maximum threshold (if data is not None).
    over : string | 'red'
        Color for values over vmax (if data is not None).
    transform : VisPy.visuals.transforms | None
        VisPy transformation to set to the parent node.
    parent : VisPy.parent | None
        Markers object parent.
    verbose : string
        Verbosity level.
    _z : float | 10.
        In case of (n_sources, 2) use _z to specify the elevation.
    kw : dict | {}
        Optional arguments are used to control the colorbar
        (See :class:`ColorbarObj`).

    Notes
    -----
    List of supported shortcuts :

        * **s** : save the figure
        * **<delete>** : reset camera

    Examples
    --------
    >>> import numpy as np
    >>> from sleep_scoring.objects import VectorObj
    >>> n_vector = 10
    >>> arrows = [np.random.rand(n_vector, 3), np.random.rand(n_vector, 3)]
    >>> data = np.random.uniform(-10, 10, (n_vector))
    >>> v = VectorObj('Vector', arrows, data=data, antialias=True)
    >>> v.preview(axis=True)
    """

    ###########################################################################
    ###########################################################################
    #                                BUILT IN
    ###########################################################################
    ###########################################################################

    def __init__(self, name, arrows, data=None, inferred_data=False,
                 select=None, color='black', dynamic=None, line_width=5.,
                 arrow_size=10., arrow_type='stealth', arrow_coef=1.,
                 arrow_norm=(5., 20.), antialias=False, cmap='viridis',
                 clim=None, vmin=None, under='gray', vmax=None, over='red',
                 transform=None, parent=None, verbose=None, _z=-10., **kw):
        """Init."""
        # Init Visbrain object base class and SourceProjection :
        VisbrainObject.__init__(self, name, parent, transform, verbose, **kw)
        self._update_cbar_args(cmap, clim, vmin, vmax, under, over)

        # _______________________ START // END _______________________
        if isinstance(arrows, (list, tuple)) and len(arrows) == 2:
            arr = list(arrows).copy()
            arrows = np.zeros(arr[0].shape[0], dtype=ARROW_DTYPES[0])
            arrows['start'] = arr[0]
            arrows['end'] = arr[1]
        assert isinstance(arrows, np.ndarray)
        # Select :
        n_arrows = len(arrows)
        select = np.ones(n_arrows, dtype=bool) if select is None else select
        arrows = arrows[select]
        data = data[select] if isinstance(data, np.ndarray) else data
        self._n_arrows = len(arrows)
        # Build (arrow_start, arrow_end)
        assert select.dtype == bool and len(select) == n_arrows
        if arrows.dtype == ARROW_DTYPES[0]:    # (start, end)
            arrow_start, arrow_end = arrows['start'], arrows['end']
            if inferred_data:
                data = np.linalg.norm(arrow_end - arrow_start, axis=1)
        elif arrows.dtype == ARROW_DTYPES[1]:  # (vertices, normals)
            assert len(arrow_norm) == 2
            norm = np.ones(n_arrows) if not isinstance(
                data, np.ndarray) else data.copy()
            norm = normalize(norm, arrow_norm[0], arrow_norm[1]).reshape(-1, 1)
            arrow_start = arrows['vertices']
            arrow_end = arrow_start + norm * arrows['normals']
        else:
            raise ValueError("Undefined type for the `arrows` input.")
        assert arrow_coef >= 1.
        arrow_end *= arrow_coef

        # _______________________ CHECKING _______________________
        # Line width // arrow type / size :
        assert isinstance(line_width, (int, float))
        assert arrow_type in ARROW_TYPES and isinstance(arrow_size, float)
        self._line_width = line_width
        self._arrow_size = arrow_size
        self._arrow_type = arrow_type
        # Get color :
        if isinstance(data, np.ndarray):
            # Clim :
            clim = (data.min(), data.max()) if clim is None else clim
            assert len(clim) == 2
            # Color :
            color = array2colormap(data, cmap=cmap, clim=clim, vmin=vmin,
                                   vmax=vmax, under=under, over=over)
            # Dynamic transparency :
            if isinstance(dynamic, (tuple, list)) and len(dynamic) == 2:
                assert all(x >= 0. and x <= 1. for x in dynamic)
                color[..., -1] = normalize(data.copy(), dynamic[0], dynamic[1])
        else:
            color = np.tile(color2vb(color).reshape(1, -1), (len(self), 1))

        # _______________________ ARROWS _______________________
        # Build arrows :
        pos = np.c_[arrow_start, arrow_end]
        line = pos.reshape(len(self) * 2, 3)
        line_color = np.repeat(color, 2, axis=0)
        self._arrows = Arrow(pos=line, color=line_color, arrows=pos,
                             arrow_type=arrow_type, arrow_size=arrow_size,
                             antialias=antialias, arrow_color=color,
                             connect='segments', width=line_width,
                             parent=self._node)

    def __len__(self):
        """Get the number of arrows."""
        return self._n_arrows

    def _get_camera(self):
        """Get the most adapted camera."""
        d_mean = self._arrows._pos.mean(0)
        dist = 1.1 * np.linalg.norm(self._arrows._pos, axis=1).max()
        return scene.cameras.TurntableCamera(center=d_mean, scale_factor=dist)

    # ----------- LINE_WIDTH -----------
    @property
    def line_width(self):
        """Get the line_width value."""
        return self._line_width

    @line_width.setter
    @wrap_properties
    def line_width(self, value):
        """Set line_width value."""
        assert isinstance(value, (int, float))
        value = max(1., value)
        self._arrows._width = value
        self._line_width = value
        self._arrows.update()

    # ----------- ARROW_TYPE -----------
    @property
    def arrow_type(self):
        """Get the arrow_type value."""
        return self._arrow_type

    @arrow_type.setter
    @wrap_properties
    def arrow_type(self, value):
        """Set arrow_type value."""
        self._arrows.arrow_type = value
        self._arrow_type = value
        self._arrows.update()

    # ----------- ARROW_SIZE -----------
    @property
    def arrow_size(self):
        """Get the arrow_size value."""
        return self._arrow_size

    @arrow_size.setter
    @wrap_properties
    def arrow_size(self, value):
        """Set arrow_size value."""
        self._arrows.arrow_size = value
        self._arrow_size = value
        self._arrows.update()


class CombineVectors(CombineObjects):
    """Combine vectors objects.

    Parameters
    ----------
    vobjs : VectorObj/list | None
        List of vector objects.
    select : string | None
        The name of the vector object to select.
    parent : VisPy.parent | None
        Images object parent.
    """

    def __init__(self, vobjs=None, select=None, parent=None):
        """Init."""
        CombineObjects.__init__(self, VectorObj, vobjs, select, parent)



"""Create a basic scene for objects."""
import sys
import logging

import numpy as np
from vispy import scene

from ..io import write_fig_canvas, dialog_save
from ..utils import color2vb, set_log_level, rotate_turntable, FixedCam
from ..visuals import CbarVisual
from ..config import CONFIG, PROFILER

logger = logging.getLogger('visbrain')


class VisbrainCanvas(object):
    """Create a canvas with an embeded axis.

    Parameters
    ----------
    axis : bool | False
        Add axis to canvas.
    title : string | None
        Title of the axis (only if axis=True).
    xlabel : string | None
        Label for the x-axis (only if axis=True).
    ylabel : string | None
        Label for the y-axis (only if axis=True).
    title_font_size : float | 15.
        Size of the title (only if axis=True).
    axis_font_size : float | 12.
        Size of x and y labels (only if axis=True).
    axis_color : array_like/string/tuple | 'black'
        Label, title, axis and border color (only if axis=True).
    tick_font_size : float | 10.
        Size of ticks for the x and y-axis (only if axis=True).
    name : string | None
        Canvas name.
    x_height_max : float | 80
        Height max of the x-axis.
    y_width_max : float | 80
        Width max of the y-axis.
    axis_label_margin : float | 50
        Margin between axis and label.
    tick_label_margin : float | 5
        Margin between ticks and label.
    rpad : float | 20.
        Right padding (space between the right border of the axis and the end
        of the canvas).
    bgcolor : array_like/tuple/string | 'black'
        Background color of the canvas.
    cargs : dict | {}
        Further arguments to pass to the SceneCanvas class.
    xargs : dict | {}
        Further arguments to pass to the AxisWidget (x-axis).
    yargs : dict | {}
        Further arguments to pass to the AxisWidget (y-axis).
    show : bool | False
        Show the canvas.
    """

    def __init__(self, axis=False, title=None, xlabel=None, ylabel=None,
                 title_font_size=15., axis_font_size=12., axis_color='black',
                 tick_font_size=10., name=None, x_height_max=80,
                 y_width_max=80, axis_label_margin=50, tick_label_margin=5,
                 rpad=20., bgcolor='white', add_cbar=False, cargs={}, xargs={},
                 yargs={}, cbargs={}, show=False, camera=None, shortcuts={}):
        """Init."""
        self._axis = axis
        self._title = title
        self._xlabel = xlabel
        self._ylabel = ylabel
        self._title_font_size = title_font_size
        self._axis_font_size = axis_font_size
        self._axis_color = axis_color
        self._tick_font_size = tick_font_size
        self._name = name
        self._x_height_max = x_height_max
        self._y_width_max = y_width_max
        self._axis_label_margin = axis_label_margin
        self._tick_label_margin = tick_label_margin
        self._bgcolor = bgcolor
        self._visible = show

        # ########################## MAIN CANVAS ##########################
        self.canvas = scene.SceneCanvas(keys='interactive', bgcolor=bgcolor,
                                        show=show, title=name, **cargs)

        # ########################## AXIS ##########################
        if axis:  # add axis to canvas
            # ----------- GRID -----------
            grid = self.canvas.central_widget.add_grid(margin=10)
            grid.spacing = 0

            # ----------- COLOR -----------
            axcol = color2vb(axis_color)
            kw = {'axis_label_margin': axis_label_margin,
                  'tick_label_margin': tick_label_margin,
                  'axis_font_size': axis_font_size, 'axis_color': axcol,
                  'tick_color': axcol, 'text_color': axcol,
                  'tick_font_size': tick_font_size}

            # ----------- TITLE -----------
            self._titleObj = scene.Label(title, color=axcol,
                                         font_size=title_font_size)
            self._titleObj.height_max = 40
            grid.add_widget(self._titleObj, row=0, col=0, col_span=2)

            # ----------- Y-AXIS -----------
            yargs.update(kw)
            self.yaxis = scene.AxisWidget(orientation='left', domain=(0, 129),
                                          axis_label=ylabel, **yargs)
            self.yaxis.width_max = y_width_max
            grid.add_widget(self.yaxis, row=1, col=0)

            # ----------- X-AXIS -----------
            xargs.update(kw)
            self.xaxis = scene.AxisWidget(orientation='bottom',
                                          axis_label=xlabel, **xargs)
            self.xaxis.height_max = x_height_max
            grid.add_widget(self.xaxis, row=2, col=1)

            # ----------- MAIN -----------
            self.wc = grid.add_view(row=1, col=1, camera=camera)
            self.grid = grid

            # ----------- LINK -----------
            self.xaxis.link_view(self.wc)
            self.yaxis.link_view(self.wc)

            # ----------- CBAR -----------
            rpad_col = 0
            if add_cbar:
                self.wc_cbar = grid.add_view(row=1, col=2)
                self.wc_cbar.width_max = 150.
                self.cbar = CbarVisual(width=.2, parent=self.wc_cbar.scene)
                rpad_col += 1

            # ----------- RIGHT PADDING -----------
            self._rpad = grid.add_widget(row=1, col=2 + rpad_col, row_span=1)
            self._rpad.width_max = rpad

        else:  # Ignore axis
            self.wc = self.canvas.central_widget.add_view(camera=camera)

    def __bool__(self):
        """Return if there's an axis attached to the canvas."""
        return self._axis

    def update(self):
        """Update canvas."""
        self.canvas.update()
        self.wc.update()
        if self._axis:
            self.xaxis.axis._update_subvisuals()
            self.yaxis.axis._update_subvisuals()
            self.grid.update()

    def set_default_state(self):
        """Set the default state of the camera."""
        self.camera.set_default_state()

    # ----------- PARENT -----------
    @property
    def parent(self):
        """Get the parent value."""
        return self.wc.scene

    # ----------- VISIBLE -----------
    @property
    def visible(self):
        """Get the visible value."""
        return self._visible

    @visible.setter
    def visible(self, value):
        """Set visible value."""
        if isinstance(value, bool):
            self._visible = value
            self.canvas.show = value
            self.update()

    # ----------- VISIBLE_AXIS -----------
    @property
    def axis(self):
        """Get the axis value."""
        return self._axis

    @axis.setter
    def axis(self, value):
        """Set axis value."""
        if self and isinstance(value, bool):
            self.xaxis.visible = value
            self.yaxis.visible = value
            self._titleObj.visible = value
            self._rpad.visible = value
            self._visible_axis = value
            self._axis = value

    # ----------- XLABEL -----------
    @property
    def xlabel(self):
        """Get the xlabel value."""
        return self._xlabel

    @xlabel.setter
    def xlabel(self, value):
        """Set xlabel value."""
        if self and isinstance(value, str):
            self.xaxis.axis.axis_label = value
            self.xaxis.axis._update_subvisuals()
            self._xlabel = value

    # ----------- YLABEL -----------
    @property
    def ylabel(self):
        """Get the ylabel value."""
        return self._ylabel

    @ylabel.setter
    def ylabel(self, value):
        """Set ylabel value."""
        if self and isinstance(value, str):
            self.yaxis.axis.axis_label = value
            self.yaxis.axis._update_subvisuals()
            self._ylabel = value

    # ----------- TITLE -----------
    @property
    def title(self):
        """Get the title value."""
        return self._title

    @title.setter
    def title(self, value):
        """Set title value."""
        if self and isinstance(value, str):
            self._titleObj.text = value
            self._titleObj.update()
            self._title = value

    # ----------- CAMERA -----------
    @property
    def camera(self):
        """Get the camera value."""
        return self.wc.camera

    @camera.setter
    def camera(self, value):
        """Set camera value."""
        self._camera = value
        # Set camera to the main widget :
        self.wc.camera = value
        # Link transformations with axis :
        if self:
            self.xaxis.link_view(self.wc)
            self.yaxis.link_view(self.wc)

    # ----------- AXIS_COLOR -----------
    @property
    def axis_color(self):
        """Get the axis_color value."""
        return self._axis_color

    @axis_color.setter
    def axis_color(self, value):
        """Set axis_color value."""
        self._axis_color = value
        col = color2vb(value)
        if self:
            # Title :
            self._titleObj._text_visual.color = col
            # X-axis :
            self.xaxis.axis.axis_color = col
            self.xaxis.axis.tick_color = col
            self.xaxis.axis._text.color = col
            self.xaxis.axis._axis_label.color = col
            # Y-axis :
            self.yaxis.axis.axis_color = col
            self.yaxis.axis.tick_color = col
            self.yaxis.axis._text.color = col
            self.yaxis.axis._axis_label.color = col
            self.update()

    # ----------- TITLE_FONT_SIZE -----------
    @property
    def title_font_size(self):
        """Get the title_font_size value."""
        return self._title_font_size

    @title_font_size.setter
    def title_font_size(self, value):
        """Set title_font_size value."""
        if self and isinstance(value, (int, float)):
            self._titleObj._text_visual.font_size = value
            self._title_font_size = value

    # ----------- LABEL_FONT_SIZE -----------
    @property
    def axis_font_size(self):
        """Get the axis_font_size value."""
        return self._axis_font_size

    @axis_font_size.setter
    def axis_font_size(self, value):
        """Set axis_font_size value."""
        if self and isinstance(value, (int, float)):
            self.xaxis.axis._axis_label.font_size = value
            self.yaxis.axis._axis_label.font_size = value
            self._axis_font_size = value

    # ----------- AXIS_FONT_SIZE -----------
    @property
    def tick_font_size(self):
        """Get the tick_font_size value."""
        return self._tick_font_size

    @tick_font_size.setter
    def tick_font_size(self, value):
        """Set tick_font_size value."""
        if self and isinstance(value, (int, float)):
            self.xaxis.axis._text.font_size = value
            self.yaxis.axis._text.font_size = value
            self._tick_font_size = value

    # ----------- BGCOLOR -----------
    @property
    def bgcolor(self):
        """Get the bgcolor value."""
        return self._bgcolor

    @bgcolor.setter
    def bgcolor(self, value):
        """Set bgcolor value."""
        self._bgcolor = value
        self.canvas.bgcolor = color2vb(value)


class SceneObj(object):
    """Create a scene and add objects to it.

    Parameters
    ----------
    bgcolor : string | 'black'
        Background color of the scene.
    show : bool | True
        Display the canvas.
    camera_state : dict | {}
        The default camera state to use.
    verbose : string
        Verbosity level.

    Notes
    -----
    List of supported shortcuts :

        * **s** : save the figure
        * **delete** : reset all views
    """

    def __init__(self, bgcolor='black', camera_state={}, verbose=None,
                 **kwargs):
        """Init."""
        set_log_level(verbose)
        logger.info("Creation of a scene")
        PROFILER('Scene creation')
        # Create the canvas and the grid :
        self.canvas = scene.SceneCanvas(keys='interactive', show=False,
                                        title='Object scene',
                                        bgcolor=color2vb(bgcolor), **kwargs)
        self._grid = self.canvas.central_widget.add_grid(margin=10)
        # Padding at (0, 0) :
        _rpad = self._grid.add_widget(row=0, col=0, row_span=1)
        _rpad.width_max = 20
        _rpad.height_max = 20
        self._grid.spacing = 0
        self._grid_desc = {}
        self._camera_state = camera_state
        # OpenGL f***** up with small objects. So we need rescale scenes :
        self._fix_gl = 100.

    def __getitem__(self, rowcol):
        """Get an element of the grid.

        Parameters
        ----------
        rowcol : tuple
            Tuple of two integers to get the subplot of the grid.
        """
        assert len(rowcol) == 2
        # Get the index of the subplot :
        idx_sub = self._grid.layout_array[rowcol[0] + 1, rowcol[1] + 1]
        # Find if there's multiple subplots with the same index :
        (i_row, i_col) = np.where(self._grid.layout_array == idx_sub)
        sl = (slice(i_row[0], i_row[-1] + 1), slice(i_col[0], i_col[-1] + 1))
        return self._grid[sl]

    def _gl_uniform_transforms(self):
        """Check that transforms are uniforms inside the grid."""
        for k in self._grid.children:  # loop over viewbox
            for i in k.children:  # loop over sub-scenes
                objs, transforms = [], []
                for j in i.children:  # loop of visuals in the scene
                    # Don't consider the camera :
                    if not isinstance(j, scene.cameras.BaseCamera):
                        objs.append(j)
                        transforms.append(j.transform)
                        # Get scaling factor :
                        sc = np.array([k.scale.sum() for k in transforms])
                        if sc.size and not len(np.unique(sc)) == 1:
                            unique_tf = transforms[sc.argmax()]
                            for k in objs:
                                k.transform = unique_tf
                            logger.debug("Standardized transformations")

    def _gl_transform(self, obj):
        """Get the transformation to apply to the object."""
        if hasattr(obj, 'mesh'):
            logger.debug("Object rescaled %s" % str([self._fix_gl] * 3))
            obj._scale = self._fix_gl
            sc = [self._fix_gl] * 3
            tf = scene.transforms.STTransform(scale=sc)
        else:
            tf = obj._node.transform
        # Add transformation to the node :
        if hasattr(obj, '_node'):  # VisbrainObject
            obj._node.transform = tf
        elif hasattr(obj, '_cnode'):  # combineObjects
            obj._cnode.transform = tf

    def add_to_subplot(self, obj, row=0, col=0, row_span=1, col_span=1,
                       title=None, title_size=12., title_color='white',
                       title_bold=True, use_this_cam=False, rotate=None,
                       zoom=None, camera_state={}, width_max=None,
                       height_max=None):
        """Add object to subplot.

        Parameters
        ----------
        obj : visbrain.object
            The visbrain object to add.
        row : int | 0
            Row location for the object.
        col : int | 0
            Columns location for the object.
        row_span : int | 1
            Number of rows to use.
        col_span : int | 1
            Number of columns to use.
        title : string | None
            Subplot title.
        title_size : float | 12.
            Title font size.
        title_color : string/tuple/array_like | 'white'
            Color of the title.
        title_bold : bool | True
            Use bold title.
        use_this_cam : bool | False
            If you add multiple objects to the same scene and if you want to
            use the camera of an object as the reference, turn this parameter
            to True.
        rotate : string | None
            Rotate the scene. Use 'top', 'bottom', 'left', 'right', 'front' or
            'back'. Only available for 3-D objects.
        zoom : float | None
            Zoom level. If zoom is in ]0, 1[, the size of the object decrease.
            If `zoom=1`, no zoom is applied. If zoom > 1., the size of the
            object increase. For example, `zoom=2` means that the displayed
            object will appear twice as large.
        camera_state : dict | {}
            Arguments to pass to the camera.
        width_max : float | None
            Maximum width of the subplot.
        height_max : float | None
            Maximum height of the subplot.
        """
        # Apply to objects (GL issue for small mesh) :
        self._gl_transform(obj)
        title = '' if not isinstance(title, str) else title
        if (row + 1, col + 1) not in self._grid_desc.keys():
            # Get the camera and add view to the grid :
            cam = obj._get_camera()
            cam.set_default_state()
            sub = self._grid.add_view(row=row + 1, col=col + 1,
                                      row_span=row_span, col_span=col_span,
                                      camera=cam)
            self._grid_desc[(row + 1, col + 1)] = len(self._grid.children)
            title_color = color2vb(title_color)
            tit = scene.visuals.Text(title, color=title_color, anchor_x='left',
                                     bold=title_bold, font_size=title_size,
                                     anchor_y='bottom')
            sub.add_subvisual(tit)
        else:
            sub = self[(row, col)]
            if use_this_cam:
                sub.camera = obj._get_camera()
                sub.camera.set_default_state()
            # For objects that have a mesh attribute, pass the camera :
            if hasattr(obj, 'mesh'):
                obj.mesh.set_camera(sub.camera)
            if title:
                logger.error("A title is already set. '%s' ignored" % title)
        # Fix the (height, width) max of the subplot :
        sub.height_max = height_max
        sub.width_max = width_max
        sub.add(obj.parent)
        # Zoom :
        if isinstance(zoom, (int, float)):
            assert zoom > 0, "`zoom` should be > 0"
            if isinstance(sub.camera, scene.cameras.TurntableCamera):
                sub.camera.scale_factor /= zoom
            elif isinstance(sub.camera, scene.cameras.PanZoomCamera) or \
                    isinstance(sub.camera, FixedCam):
                r = sub.camera.rect
                prop = np.array((r.width, r.height)) / zoom
                left = r.center[0] - (prop[0] / 2.)
                bottom = r.center[1] - (prop[1] / 2.)
                sub.camera.rect = (left, bottom, prop[0], prop[1])
        # Camera :
        if camera_state == {}:
            camera_state = self._camera_state
        if isinstance(sub.camera, scene.cameras.TurntableCamera):
            rotate_turntable(fixed=rotate, camera_state=camera_state,
                             camera=sub.camera)
        sub.camera.set_default_state()
        PROFILER('%s added to the scene' % repr(obj))
        logger.info('    %s added to the scene' % repr(obj))

    def link(self, *args):
        """Link the camera of several objects of the scene.

        Parameters
        ----------
        args : list
            List of tuple describing subplot locations. Alternatively, use `-1`
            to link all cameras.

        Examples
        --------
        >>> # Link cameras of subplots (0, 0), (0, 1) and (1, 0)
        >>> sc.link((0, 0), (0, 1), (1, 0))
        """
        logger.info('    Link cameras')
        if args[0] == -1:
            args = [(k[0] - 1, k[1] - 1) for k in self._grid_desc.keys()]
        assert len(args) > 1
        assert all([len(k) == 2 for k in args])
        cam_obj_1 = self[args[0]].camera
        for obj in args[1::]:
            cam_obj_1.link(self[obj].camera)
        PROFILER('Link cameras %s' % str(args))

    def screenshot(self, saveas, print_size=None, dpi=300.,
                   unit='centimeter', factor=None, region=None, autocrop=False,
                   bgcolor=None, transparent=False, line_width=1.):
        """Take a screeshot of the scene.

        By default, the rendered canvas will have the size of your screen.
        The screenshot() method provides two ways to increase to exported image
        resolution :

            * Using print_size, unit and dpi inputs : specify the size of the
              image at a specific dpi level. For example, you might want to
              have an (10cm, 15cm) image at 300 dpi.
            * Using the factor input : multiply the default image size by this
              factor. For example, if you have a (1920, 1080) monitor and if
              factor is 2, the exported image should have a shape of
              (3840, 2160) pixels.

        Parameters
        ----------
        saveas : str
            The name of the file to be saved. This file must contains a
            extension like .png, .tiff, .jpg...
        print_size : tuple | None
            The desired print size. This argument should be used in association
            with the dpi and unit inputs. print_size describe should be a tuple
            of two floats describing (width, height) of the exported image for
            a specific dpi level. The final image might not have the exact
            desired size but will try instead to find a compromize
            regarding to the proportion of width/height of the original image.
        dpi : float | 300.
            Dots per inch for printing the image.
        unit : {'centimeter', 'millimeter', 'pixel', 'inch'}
            Unit of the printed size.
        factor : float | None
            If you don't want to use the print_size input, factor simply
            multiply the resolution of your screen.
        region : tuple | None
            Select a specific region. Must be a tuple of four integers each one
            describing (x_start, y_start, width, height).
        autocrop : bool | False
            Automaticaly crop the figure in order to have the smallest
            space between the brain and the border of the picture.
        bgcolor : array_like/string | None
            The background color of the image.
        transparent : bool | False
            Specify if the exported figure have to contains a transparent
            background.
        """
        kwargs = dict(print_size=print_size, dpi=dpi, factor=factor,
                      autocrop=autocrop, unit=unit, region=region,
                      bgcolor=bgcolor, transparent=transparent)
        self._gl_uniform_transforms()
        write_fig_canvas(saveas, self.canvas,
                         widget=self.canvas.central_widget, **kwargs)

    def _scene_shortcuts(self):
        """Add shortcuts to the scene."""
        # On key pressed :
        def key_pressed(event):  # noqa
            if event.text == 's':
                from PyQt5.QtWidgets import QWidget
                ext = ['png', 'tiff', 'jpg']
                _ext = ['%s file (*.%s)' % (k.upper(), k) for k in ext]
                _ext += ['All files (*.*)']
                saveas = dialog_save(QWidget(), name='Export the scene',
                                     default='canvas.png', allext=_ext)
                if saveas:
                    write_fig_canvas(saveas, self.canvas,
                                     widget=self.canvas.central_widget)
        self.canvas.events.key_press.connect(key_pressed)

    def preview(self):
        """Previsualize the result."""
        self._gl_uniform_transforms()
        self.canvas.show(visible=True)
        # Shortcuts :
        self._scene_shortcuts()
        # Profiler :
        if PROFILER and logger.level == 1:
            logger.profiler("PARENT TREE \n%s" % self._grid.describe_tree())
            logger.profiler(" ")
            PROFILER.finish()
        if sys.flags.interactive != 1 and CONFIG['SHOW_PYQT_APP']:
            CONFIG['VISPY_APP'].run()



"""Time-frequency map object."""
import logging
import numpy as np
from scipy.signal import spectrogram

from .image_obj import ImageObj
from ..utils import (morlet, averaging, normalization)
from ..io.dependencies import is_lspopt_installed

logger = logging.getLogger('visbrain')


class TimeFrequencyObj(ImageObj):
    """Compute the time-frequency map (or spectrogram).

    The time-frequency decomposition can be assessed using :

        * The fourier transform
        * Morlet's wavelet
        * Multi-taper

    Parameters
    ----------
    name : string | None
        Name of the time-frequency object.
    data : array_like
        Array of data of shape (N,)
    sf : float | 1.
        The sampling frequency.
    method : {'fourier', 'wavelet', 'multitaper'}
        The method to use to compute the time-frequency decomposition.
    nperseg : int | 256
        Length of each segment. Argument pass to the `scipy.signal.spectrogram`
        function (for 'fourier' and 'multitaper' method).
    overlap : float | 0.
        Overlap between segments. Must be between 0. and 1.
    f_min : float | 1.
        Minimum frequency (for 'wavelet' method).
    f_max : float | 160.
        Maximum frequency (for 'wavelet' method).
    f_step : float | 2.
        Frequency step between two consecutive frequencies (for 'wavelet'
        method).
    baseline : array_like | None
        Baseline period (for 'wavelet' method).
    norm : int | None
        The normalization type (for 'wavelet' method).. See the `normalization`
        function.
    n_window : int | None
        If this parameter is an integer, the time-frequency map is going to be
        averaged into smaller windows (for 'wavelet' method).
    window : {'flat', 'hanning', 'hamming', 'bartlett', 'blackman'}
        Windowing method for averaging. By default, 'flat' is used for Wavelet
        and 'hamming' for Fourier.
    c_parameter : int | 20
        Parameter 'c' described in doi:10.1155/2011/980805 (for 'multitaper'
        method)
    clim : tuple | None
        Colorbar limits. If None, `clim=(data.min(), data.max())`
    cmap : string | None
        Colormap name.
    vmin : float | None
        Minimum threshold of the colorbar.
    under : string/tuple/array_like | None
        Color for values under vmin.
    vmax : float | None
        Maximum threshold of the colorbar.
    over : string/tuple/array_like | None
        Color for values over vmax.
    interpolation : string | 'nearest'
        Interpolation method for the image. See vispy.scene.visuals.Image for
        availables interpolation methods.
    max_pts : int | -1
        Maximum number of points of the image along the x or y axis. This
        parameter is essentially used to solve OpenGL issues with very large
        images.
    transform : VisPy.visuals.transforms | None
        VisPy transformation to set to the parent node.
    parent : VisPy.parent | None
        Markers object parent.
    verbose : string
        Verbosity level.
    kw : dict | {}
        Optional arguments are used to control the colorbar
        (See :class:`ColorbarObj`).

    Notes
    -----
    List of supported shortcuts :

        * **s** : save the figure
        * **<delete>** : reset camera

    Examples
    --------
    >>> import numpy as np
    >>> from sleep_scoring.objects import TimeFrequencyObj
    >>> n, sf = 512, 256  # number of time-points and sampling frequency
    >>> time = np.arange(n) / sf  # time vector
    >>> data = np.sin(2 * np.pi * 25. * time) + np.random.rand(n)
    >>> tf = TimeFrequencyObj('tf', data, sf)
    >>> tf.preview(axis=True)
    """

    def __init__(self, name, data=None, sf=1., method='fourier', nperseg=256,
                 f_min=1., f_max=160., f_step=1., baseline=None, norm=None,
                 n_window=None, overlap=0., window=None, c_parameter=20,
                 cmap='viridis', clim=None, vmin=None, under='gray', vmax=None,
                 over='red', interpolation='nearest', max_pts=-1, parent=None,
                 transform=None, verbose=None, **kw):
        """Init."""
        # Initialize the image object :
        ImageObj.__init__(self, name, interpolation=interpolation,
                          max_pts=max_pts, parent=parent, transform=transform,
                          verbose=verbose, **kw)

        # Compute TF and set data to the ImageObj :
        if isinstance(data, np.ndarray):
            self.set_data(data, sf, method, nperseg, f_min, f_max, f_step,
                          baseline, norm, n_window, overlap, window,
                          c_parameter, clim, cmap, vmin, under, vmax, over)

    def set_data(self, data, sf=1., method='fourier', nperseg=256, f_min=1.,
                 f_max=160., f_step=1., baseline=None, norm=None,
                 n_window=None, overlap=0., window=None, c_parameter=20,
                 clim=None, cmap='viridis', vmin=None, under=None, vmax=None,
                 over=None):
        """Compute TF and set data to the ImageObj."""
        # ======================= CHECKING =======================
        assert isinstance(data, np.ndarray) and data.ndim == 1
        assert isinstance(sf, (int, float))
        assert method in ('fourier', 'wavelet', 'multitaper')
        if not isinstance(window, str):
            window = 'hamming' if method is 'fourier' else 'flat'
        assert 0. <= overlap < 1.
        # Wavelet args :
        assert isinstance(f_min, (int, float))
        assert isinstance(f_max, (int, float))
        assert isinstance(f_step, (int, float))
        # Spectrogram and Multi-taper args :
        noverlap = int(round(overlap * nperseg))
        assert isinstance(nperseg, int)
        assert isinstance(c_parameter, int)

        # Update color arguments :
        self._update_cbar_args(cmap, clim, vmin, vmax, under, over)
        logger.info("    Compute time-frequency decomposition using the"
                    " %s method" % method)

        if method == 'fourier':
            freqs, time, tf = spectrogram(data, sf, nperseg=nperseg,
                                          noverlap=noverlap, window=window)
        if method == 'wavelet':
            n_pts = len(data)
            freqs = np.arange(f_min, f_max, f_step)
            time = np.arange(n_pts) / sf
            tf = np.zeros((len(freqs), n_pts), dtype=data.dtype)
            # Compute TF and inplace normalization :
            logger.info("    Compute the time-frequency map ("
                        "normalization=%r)" % norm)
            for i, k in enumerate(freqs):
                tf[i, :] = np.square(np.abs(morlet(data, sf, k)))
            normalization(tf, norm=norm, baseline=baseline, axis=1)

            # Averaging :
            if isinstance(n_window, int):
                logger.info("    Averaging time-frequency map using windows of"
                            " size %i with a %f overlap" % (n_window, overlap))
                kw = dict(overlap=overlap, window=window)
                tf = averaging(tf, n_window, axis=1, **kw)
                time = averaging(time, n_window, **kw)
        elif method == 'multitaper':
            is_lspopt_installed(raise_error=True)
            from lspopt import spectrogram_lspopt
            freqs, time, tf = spectrogram_lspopt(data, sf, nperseg=nperseg,
                                                 noverlap=noverlap,
                                                 c_parameter=c_parameter)

        # Set data to the image object :
        ImageObj.set_data(self, tf, xaxis=time, yaxis=freqs,
                          **self.to_kwargs())



"""This script contains some other utility functions."""
import numpy as np
from vispy.visuals.transforms import (STTransform, ChainTransform,
                                      MatrixTransform)


__all__ = ('vprescale', 'vprecenter', 'vpnormalize', 'array_to_stt',
           'stt_to_array')


def vprescale(obj, dist=1.):
    """Get the vispy transformation for rescaling objects.

    Parameters
    ----------
    obj : tuple
        The ndarray of the object to rescale. Must be a (..., 2) or (..., 3).
    dist : float | 1.
        The final rescaling value.

    Returns
    -------
    rescale : vispy.transformations
        VisPy transformation to rescale.
    """
    # Get minimum / maximum trough last dimension :
    dim = tuple(np.arange(obj.ndim - 1, dtype=int))
    ptp = np.max(obj, axis=dim) - np.min(obj, axis=dim)
    return STTransform(scale=[dist / np.max(ptp)] * 3)


def vprecenter(obj):
    """Get the vispy transformation to recenter objects.

    Parameters
    ----------
    obj : tuple
        The ndarray of the object to recenter. Must be a (..., 2) or (..., 3).

    Returns
    -------
    recenter : vispy.transformations
        VisPy transformation to recenter.
    """
    # Check object :
    if not isinstance(obj, np.ndarray) or obj.shape[-1] not in [2, 3]:
        raise ValueError("The object must be a ndarray of shape (..., 2) or "
                         "(..., 3)")
    # Get center :
    center = np.mean(obj, axis=tuple(np.arange(obj.ndim - 1, dtype=int)))
    if center.shape[-1] == 2:
        center = np.hstack((center, 0.))
    # Build transformation :
    return STTransform(translate=-center)


def vpnormalize(obj, dist=1.):
    """Get the vispy transformation for normalizing objects.

    Parameters
    ----------
    obj : tuple
        The ndarray of the object to rescale. Must be a (..., 2) or (..., 3).
    dist : float | 1.
        The final rescaling value.

    Returns
    -------
    normalize : vispy.transformations
        VisPy transformation to normalize.
    """
    # Prepare the transformation chain :
    t = ChainTransform()
    # Recenter the object :
    t.prepend(vprecenter(obj))
    # Rescale :
    t.prepend(vprescale(obj, dist))
    return t


def array_to_stt(arr):
    """Turn a (4, 4) array into a scale and translate matrix transformation.

    Parameters
    ----------
    arr : array_like
        A (4, 4) array.

    Returns
    -------
    transform : VisPy.transform
        The VisPy transformation.
    """
    assert isinstance(arr, np.ndarray) and arr.shape == (4, 4)
    _arr = arr.copy()
    _arr[-1, 0:-1] = _arr[0:-1, -1]
    _arr[0:-1, -1] = 0.
    transform = MatrixTransform(_arr)
    # transform.scale(np.diag(_arr)[0:-1])
    # transform.translate(_arr[0:-1, -1])
    return transform


def stt_to_array(stt):
    """Convert a MatrixTransform into a (4, 4) array.

    Parameters
    ----------
    stt : MatrixTransform
        The transformation.

    Returns
    -------
    mat : array_like
        The (4, 4) array.
    """
    assert isinstance(stt, MatrixTransform)
    mat = stt.matrix.copy()
    mat[0:-1, -1] = mat[-1, 0:-1]
    mat[-1, 0:-1] = 0.
    return mat



"""This file contains several functions for memory usage.

Taken from the numpy tricks : http://ipython-books.github.io/featured-01/
"""
import numpy as np


__all__ = ('id', 'arrays_share_data', 'code_timer')


def id(x):
    """Get the memory block address of an array."""
    return x.__array_interface__['data'][0]


def get_data_base(arr):
    """For a given array, finds the base array that "owns" the actual data."""
    base = arr
    while isinstance(base.base, np.ndarray):
        base = base.base
    return base


def arrays_share_data(x, y):
    """Return if two arrays share an offset."""
    return get_data_base(x) is get_data_base(y)


def code_timer(previous=0., verbose=True, prefix='', unit='s'):
    """Time code execution.

    Parameters
    ----------
    previous : float | 0.
        Previous code timing.
    verbose : bool | True
        Print time difference.
    prefix : string | ''
        Prefix to add before printing.
    unit : string | 's'
        Use either 's', 'ms' 'us'.
    """
    from time import time
    assert unit in ['s', 'ms', 'us']

    fact = 1. if unit == 's' else 10e3 if unit == 'ms' else 10e6
    current = time()
    st = abs(current - previous)
    if verbose:
        print(prefix, st * fact, '(' + unit + ')')
    return current



"""Usefull functions for graphical interface managment."""

from PyQt5 import QtCore

import numpy as np

from .color import color2vb, color2tuple

__all__ = ('slider2opacity', 'textline2color', 'color2json',
           'ndsubplot', 'combo', 'is_color', 'MouseEventControl',
           'disconnect_all', 'extend_combo_list', 'get_combo_list_index',
           'safely_set_cbox', 'safely_set_spin', 'safely_set_slider',
           'toggle_enable_tab', 'get_screen_size', 'set_widget_size',
           'fill_pyqt_table')


def slider2opacity(value, thmin=0.0, thmax=100.0, vmin=-5.0, vmax=105.0,
                   tomin=-1000.0, tomax=1000.0):
    """Convert a slider value into opacity.

    Parameters
    ----------
    value : float
        The slider value
    thmin : float | 0.0
        Minimum threshold to consider
    thmax : float | 100.0
        Maximum threshold to consider
    tomin : float | -1000.0
        Set to tomin if value is under vmin
    tomax : float | 1000.0
        Set to tomax if value is over vmax

    Returns
    -------
    color : array_like
        Array of RGBA colors
    """
    alpha = 0.
    # Linear decrease :
    if value < thmin:
        alpha = value * tomin / vmin
    # Linear increase :
    elif value > thmax:
        alpha = value * tomax / vmax
    else:
        alpha = value / 100
    return alpha


def textline2color(value):
    """Transform a Qt text line editor into color.

    Parameters
    ----------
    value : string
        The edited string

    Returns
    -------
    value : string
        The processed string value.
    color : tuple
        Tuple of RGBA colors
    """
    # Remove ' caracter :
    try:
        value = value.replace("'", '').strip()
        # Tuple/list :
        try:
            if isinstance(eval(value), (tuple, list)):
                value = eval(value)
        except:
            pass
        return value, color2vb(color=value)
    except:
        return 'white', (1., 1., 1., 1.)


def color2json(obj, rmalpha=True):
    """Turn a color textline into a json compatible one.

    Parameters
    ----------
    obj : PyQt textline object
        The PyQt text line object.
    rmalpha : bool | True
        Specify if the alpha component have to be deleted.

    Returns
    -------
    coltuple : tuple
        A json compatible tuple of floating points.
    """
    return color2tuple(textline2color(obj.text())[1], float, rmalpha)


def is_color(color, comefrom='color'):
    """Test if a variable is a color.

    Parameters
    ----------
    color : str/tuple/list/array
        The color to test.
    comefrom : string | 'color'
        Where come from the color object. Use either 'color' if it has to
        be considered directly as a color or 'textline' if it comes from a
        textline gui objects.

    Returns
    -------
    iscol : bool
        A boolean value to indicate if it is a color.
    """
    if comefrom == 'color':
        try:
            color2vb(color)
            iscol = True
        except:
            iscol = False
    elif comefrom == 'textline':
        try:
            color = color.replace("'", '')
            try:
                if isinstance(eval(color), (tuple, list)):
                    color = eval(color)
            except:
                pass
            color2vb(color=color)
            iscol = True
        except:
            iscol = False
    else:
        raise ValueError("The comefrom must either be 'color' or 'textline'.")

    return iscol


def ndsubplot(n, line=4, force_col=None, max_rows=100, max_on_line=True):
    """Get the optimal number of rows / columns for a given integer.

    Parameters
    ----------
    n : int
        The length to convert into rows / columns.
    line : int | 4
        If n <= line, the number of rows will be forced to be 1.
    force_col : int | None
        Force the number of columns.
    max_rows : int | 10
        Maximum number of rows.
    max_on_line : bool | True
        Force to have the maximum set for lines.

    Returns
    -------
    nrows : int
        The number of rows.
    ncols : int
        The number of columns.
    """
    # Force n to be integer :
    n = int(n)
    # Force to have a single line subplot :
    if n <= line:
        ncols, nrows = n, 1
    else:
        if force_col is not None:
            ncols = force_col
            nrows = int(n / ncols)
        else:
            # Build a linearly decreasing vector :
            vec = np.linspace(max_rows, 2, max_rows + 1,
                              endpoint=False).astype(int)
            # Compute n modulo each point in vec :
            mod, div = n % vec, n / vec
            # Find where the result is zero :
            nbool = np.invert(mod.astype(bool))
            if any(nbool):
                cmin = np.abs(vec[nbool] - div[nbool]).argmin()
                ncols = vec[nbool][cmin]
                nrows = int(n / ncols)
            else:
                nrows, ncols = 1, n

    if max_on_line and nrows < ncols:
        nrows, ncols = (ncols, nrows)

    return nrows, ncols


def combo(lst, idx):
    """Manage combo box.

    Parameters
    ----------
    lst : list
        List of possible values.
    idx : list
        List of index of several combo box.

    Returns
    -------
    out : list
        List of possible values for each combo box.
    ind : list
        List of the new current index of each combo box.
    """
    out, ind, original = [], [], set(lst)
    for k in range(len(idx)):
        out.append(list(original.difference(idx[:k])))
        # ind.append(lst.index(idx[k]))
        ind.append(list(out)[k][0])
        # ind.append(out[k].index(idx[k]))
    return out, ind


class MouseEventControl(object):
    """Additional mouse control on VisPy canvas."""

    def _is_left_click(self, event):
        """Return if the pressed button is the left one."""
        return event.button == 1

    def _is_modifier(self, event, modifier):
        """Return the name of the modifier use."""
        try:
            return event.modifiers[0].name == modifier
        except:
            return False


def disconnect_all(obj):
    """Disconnect all functions related to an PyQt object.

    Parameters
    ----------
    obj : PyQt object
        The PyQt object to disconnect.
    """
    while True:
        try:
            obj.disconnect()
        except TypeError:
            break


def extend_combo_list(cbox, item, reconnect=None):
    """Extend a QtComboList with a new item.

    Parameters
    ----------
    cbox : PyQt.QtComboList
        The PyQt combo list object.
    item : string
        Name of the new item.
    reconnect : function | None
        The function to apply when the index changed.
    """
    # Get the list of current items and extend it :
    all_items = [cbox.itemText(i) for i in range(cbox.count())]
    all_items.append(item)
    # Reconnect function :
    is_connected = reconnect is not None
    # Disconnect if connected :
    if is_connected:
        cbox.disconnect()
    # Clear and safely add items :
    cbox.clear()
    cbox.addItems(all_items)
    # Reconnect if connected :
    if is_connected:
        cbox.currentIndexChanged.connect(reconnect)


def get_combo_list_index(cbox, name):
    """Get index of an item in a combo box.

    Parameters
    ----------
    cbox : PyQt.QtComboList
        The PyQt combo list object.
    name : string
        Name of the item.

    Returns
    -------
    index : int
        Index of the item in the combo box.
    """
    # Get the list of current items and extend it :
    all_items = [cbox.itemText(i) for i in range(cbox.count())]
    return all_items.index(name)


def safely_set_cbox(cbox, idx, fcn=None):
    """Set QtComboBox list index without trigger.

    Parameters
    ----------
    cbox : PyQt.QtComboList
        The PyQt combo list object.
    idx : float/string
        Index or name of the item.
    fcn : list | None
        List of functions to disconnect then reconnect.
    """
    if isinstance(fcn, list):
        disconnect_all(cbox)
    if isinstance(idx, str):
        idx = get_combo_list_index(cbox, idx)
    cbox.setCurrentIndex(idx)
    if isinstance(fcn, list):
        for k in fcn:
            cbox.currentIndexChanged.connect(k)


def safely_set_spin(spin, value, fcn, keyboard_tracking=True):
    """Set value to QtSpin without trigger.

    Parameters
    ----------
    spin : QtSpin
        The Qt spin object.
    value : float
        Value to set.
    fcn : function
        List of function to reconnect.
    keyboard_tracking : bool | True
        Trigger while pressing values on keyboard.
    """
    # Disconnect and set value :
    disconnect_all(spin)
    spin.setValue(value)
    # Reconnect :
    for k in fcn:
        spin.valueChanged.connect(k)
    spin.setKeyboardTracking(False)


def safely_set_slider(slider, value, fcn):
    """Set value to QtSlider without trigger.

    Parameters
    ----------
    slider : QtSlider
        The Qt slider object.
    value : float
        Value to set.
    fcn : function
        List of function to reconnect.
    """
    # Disconnect and set value :
    disconnect_all(slider)
    slider.setValue(value)
    # Reconnect :
    for k in fcn:
        slider.sliderMoved.connect(k)


def toggle_enable_tab(tab, name, enable=False):
    """Enable or disable a tab based on the name.

    Parameters
    ----------
    tab : PyQt.QTabWidget
        The PyQt tab.
    name : string
        Name of the tab.
    enable : bool | False
        Enable or disble the tab.
    """
    # Get all tab names :
    names = [tab.tabText(k) for k in range(tab.count())]
    # Get index of named tab :
    idx = names.index(name)
    # Set tab enable/disable :
    tab.setTabEnabled(idx, enable)


def get_screen_size(app):
    """Get screen size of an application.

    Parameters
    ----------
    app : QtApplication
        A PyQt application.

    Returns
    -------
    width : int
        Width of the application.
    height : int
        Height of the application.
    """
    resolution = app.desktop().screenGeometry()
    return resolution.width(), resolution.height()


def set_widget_size(app, widget, width=100., height=100.):
    """Set widget size proportionaly to screen resolution.

    Parameters
    ----------
    app : QtApplication
        A PyQt application.
    widget : QtWidget
        The PyQt widget.
    width : float | 100.
        Proportional width (0 < width <= 100).
    height : float | 100.
        Proportional height (0 < height <= 100).
    """
    # Check width and height :
    if not 0. < width <= 100.:
        raise ValueError("The width parameter must be 0 < width <= 100")
    if not 0. < height <= 100.:
        raise ValueError("The height parameter must be 0 < height <= 100")
    # Get scren (width, height) :
    s_width, s_height = get_screen_size(app)
    # Convert (width, height) into pixels :
    # TODO: casted these into ints
    s_width = int(np.around(s_width * width / 100.))
    s_height = int(np.around(s_height * height / 100.))
    # Set maximum size to the widget :
    size = QtCore.QSize(s_width, s_height)
    widget.resize(size)


def fill_pyqt_table(table, col_names=None, col=None, df=None, filter=None,
                    filter_col=0, check=None):
    """Fill a PyQt table widget.

    Parameters
    ----------
    col_names : list | None
        List of name of each columns.
    col : list | None
        List of columns values.
    df : pandas.DataFrame or dict | None
        Alternatively, a pandas DataFrame or a dictionary can also be used.
    """
    from PyQt5.QtWidgets import (QTableWidgetItem, QTableWidget, QTableView)
    from PyQt5 import QtGui, QtCore

    # ________________________ Checking ________________________
    # Dictionary / pandas.DataFrame :
    if df is not None:
        col_names = list(df.keys())
        col = []
        for k in col_names:
            col.append(df[k])
    assert len(col_names) == len(col)
    assert all([isinstance(k, str) for k in col_names])
    n_rows, n_cols = (len(col[0]), len(col_names))

    # Switch between table view / widget :
    if isinstance(table, QTableWidget):  # Table widget
        table.clear()
        table.setColumnCount(n_cols)
        table.setHorizontalHeaderLabels(col_names)
        table.setRowCount(n_rows)

        for i in range(table.rowCount()):
            for k in range(table.columnCount()):
                table.setItem(i, k, QTableWidgetItem(str(col[k][i])))
    elif isinstance(table, QTableView):  # Table view
        table.reset()
        # Fill the model :
        model = QtGui.QStandardItemModel(n_rows, n_cols)
        model.setHorizontalHeaderLabels(col_names)
        for i in range(n_rows):
            for k in range(n_cols):
                item = QtGui.QStandardItem(str(col[k][i]))
                item.setCheckable(check == k)
                model.setItem(i, k, item)
        table.setModel(model)

        if filter is not None:
            filt_model = QtCore.QSortFilterProxyModel()
            filt_model.setSourceModel(model)
            filt_model.setFilterKeyColumn(filter_col)
            filt_model.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)
            filter.textChanged.connect(filt_model.setFilterRegExp)
            table.setModel(filt_model)
        return model



"""This script contains some other utility functions."""
import logging

import numpy as np
from vispy.util import profiler


__all__ = ('Profiler', 'get_dsf', 'set_if_not_none')


class Profiler(object):
    """Visbrain profiler.

    The visbrain profiler add some basic functionalities to the vispy profiler.
    """

    def __init__(self, delayed=True):
        """Init."""
        self._delayed = delayed
        logger = logging.getLogger('visbrain')
        enable = logger.level == 1  # enable for PROFILER
        if enable and not hasattr(self, '_vp_profiler'):
            self._vp_profiler = profiler.Profiler(disabled=not enable,
                                                  delayed=self._delayed)

    def __bool__(self):
        """Return if the profiler is enable."""
        if hasattr(self, '_vp_profiler'):
            return not isinstance(self._vp_profiler,
                                  profiler.Profiler.DisabledProfiler)
        else:
            return False

    def __call__(self, msg=None, level=0, as_type='msg'):
        """Call the vispy profiler."""
        self.__init__(delayed=self._delayed)
        if self:
            if as_type == 'msg':
                if isinstance(msg, str) and isinstance(level, int):
                    msg = '    ' * level + '> ' + msg
                self._vp_profiler(self._new_msg(msg))
            elif as_type == 'title':
                depth = type(self._vp_profiler)._depth
                msg = "  " * depth + '-' * 6 + ' ' + msg + ' ' + '-' * 6
                self._vp_profiler._new_msg(self._new_msg(msg))

    def finish(self, msg=None):
        """Finish the profiler."""
        self._vp_profiler.finish(msg)

    @staticmethod
    def _new_msg(msg):
        msg += ' ' if msg[-1] != ' ' else ''
        return msg


def get_dsf(downsample, sf):
    """Get the downsampling factor.

    Parameters
    ----------
    downsample : float
        The down-sampling frequency.
    sf : float
        The sampling frequency
    """
    if all([isinstance(k, (int, float)) for k in (downsample, sf)]):
        dsf = int(np.round(sf / downsample))
        downsample = float(sf / dsf)
        return dsf, downsample
    else:
        return 1, downsample


def set_if_not_none(to_set, value, cond=True):
    """Set a variable if the value is not None.

    Parameters
    ----------
    to_set : string
        The variable name.
    value : any
        The value to set.
    cond : bool | True
        Additional condition.

    Returns
    -------
    val : any
        The value if not None else to_set
    """
    return value if (value is not None) and cond else to_set



"""Plot a forward solution.

Annalisa Pascarella
"""
import numpy as np
import logging
import os

from ..io import is_mne_installed

logger = logging.getLogger('visbrain')

__all__ = ['mne_plot_source_estimation', 'mne_plot_source_space']


def _extract_arrays_from_src(src, hemisphere='both', fact=1.):
    """Get vertices, faces, active vertices and sources from SourceSpace."""
    logger.debug("Loading vert, faces, activ and sources from src structure.")
    # Define usefull variables :
    _l_nb, _r_nb = src[0]['rr'].shape[0], src[1]['rr'].shape[0]
    _f_off, _act_off = [0, src[0]['tris'].max() + 1], [0, _l_nb]
    _hemi = {'left': [0], 'right': [1], 'both': [0, 1]}[hemisphere]
    # Get vertices and faces :
    vertices = np.vstack([src[k]['rr'] for k in [0, 1]])
    faces = np.vstack([src[k]['tris'] + _f_off[k] for k in [0, 1]])
    lr_index = np.r_[np.ones((_l_nb,)), np.zeros((_r_nb))].astype(bool)
    # Get active vertex and sources :
    activ = np.hstack([src[k]['vertno'] + _act_off[k] for k in [0, 1]])
    sources = np.vstack([src[k]['rr'][src[k]['vertno']] for k in _hemi])
    return fact * vertices, faces, lr_index, activ, fact * sources


def _plt_src(name, kw_brain_obj, active_data, active_vert, sources,
             kw_source_obj, kw_activation, show):
    # Define a brain object and a source object :
    logger.info('    Define a Brain and Source objects')
    from sleep_scoring.objects import BrainObj, SourceObj, SceneObj
    brain_obj, source_obj = name + '_brain', name + '_sources'
    b_obj = BrainObj(brain_obj, **kw_brain_obj)
    s_obj = SourceObj(source_obj, sources, **kw_source_obj)
    s_obj.visible_obj = False
    # Add data to the BrainObj if needed :
    if isinstance(active_data, np.ndarray):
        logger.info("    Add active data between "
                    "[%2f, %2f]" % (active_data.min(), active_data.max()))
        b_obj.add_activation(data=active_data, vertices=active_vert,
                             **kw_activation)
    # Return either a scene or a BrainObj and SourceObj :
    if show is True:  # Display inside the Brain GUI
        # Define a Brain instance :
        from sleep_scoring.gui import Brain
        brain = Brain(brain_obj=b_obj, source_obj=s_obj)
        brain._brain_template.setEnabled(False)
        # By default, display colorbar if activation :
        if isinstance(active_data, np.ndarray):
            brain.menuDispCbar.setChecked(True)
            brain._fcn_menu_disp_cbar()
        brain.show()
    elif show is 'scene':  # return a SceneObj
        logger.info("    Define a unique scene for the Brain and Source "
                    "objects")
        sc = SceneObj()
        sc.add_to_subplot(s_obj)
        sc.add_to_subplot(b_obj, use_this_cam=True)
        return sc
    else:  # return the BrainObj and SourceObj
        s_obj.visible_obj = True
        return b_obj, s_obj


def mne_plot_source_estimation(sbj, sbj_dir, fwd_file, stc_file=None,
                               hemisphere='both', parc='aparc', active_data=0,
                               kw_brain_obj={}, kw_source_obj={},
                               kw_activation={}, show=True):
    """Plot source estimation.

    Parameters
    ----------
    sbj : string
        The subject name.
    sbj_dir : string
        Path to the subject directory.
    fwd_file : string
        The file name of the forward solution, which should end with -fwd.fif
        or -fwd.fif.gz.
    stc_file : string | None
        Path to the .stc inverse solution file.
    hemisphere : {'left', 'both', 'right'}
        The hemisphere to plot.
    parc : string | 'aparc'
        The parcellation to use, e.g., ‘aparc’ or ‘aparc.a2009s’.
    active_data : array_like, int | 0
        The data to set to vertices. If an stc file is provided and if
        `active_data` is an integer, it describes the time instant in which you
        want to see the activation. Otherwise, `active_data` must be an array
        with the same same shape as the number of active vertices.
    kw_brain_obj : dict | {}
        Additional inputs to pass to the :class:`visbrain.objects.BrainObj`
        class.
    kw_source_obj : dict | {}
        Additional inputs to pass to the :class:`visbrain.objects.SourceObj`
        class.
    kw_activation : dict | {}
        Additional inputs to pass to the
        :class:`visbrain.objects.BrainObj.add_activation` method.
    show : bool | False
        If True, the window of the :class:`visbrain.Brain` module is
        automatically displayed. If False, a :class:`visbrain.objects.BrainObj`
        and a :class:`visbrain.objects.SourceObj` are returned. Finally, if
        'scene'a :class:`visbrain.objects.SceneObj` is returned.

    Returns
    -------
    b_obj : BrainObj
        A predefined :class:`visbrain.objects.BrainObj` (if `show=False`)
    s_obj : SourceObj
        A predefined :class:`visbrain.objects.SourceObj`, hide by default (if
        `show=False`)
    """
    # Test that mne is installed and import :
    is_mne_installed(raise_error=True)
    import mne
    from mne.source_space import head_to_mni
    # Read the forward solution :
    fwd = mne.read_forward_solution(fwd_file)
    logger.debug('Read the forward solution')
    # Get source space :
    fwd_src = fwd['src']
    # Get the MRI (surface RAS)-> head matrix
    mri_head_t = fwd['mri_head_t']
    # Extract arrays from src structure :
    (vertices, faces, lr_index, active_vert,
     sources) = _extract_arrays_from_src(fwd_src, hemisphere)
    # Head to MNI conversion
    logger.info("    Head to MNI conversion")
    vertices = head_to_mni(vertices, sbj, mri_head_t, subjects_dir=sbj_dir)
    sources = head_to_mni(sources, sbj, mri_head_t, subjects_dir=sbj_dir)
    # Add data to the mesh :
    if isinstance(active_data, np.ndarray):
        if len(active_data) != len(active_vert):
            logger.error("The length of `active data` (%i) must be the same "
                         "the length of the number of active vertices "
                         "(%i)" % (len(active_data), len(active_vert)))
            active_data = active_vert = None
        else:
            logger.info("    Array of active data used.")
    elif isinstance(stc_file, str) and isinstance(active_data, int):
        # Get active data :
        assert os.path.isfile(stc_file)
        n_tp = active_data
        data = mne.read_source_estimate(stc_file).data
        active_data = np.abs(data[:, n_tp] / data[:, n_tp].max())
        logger.info("    Time instant %i used for activation" % n_tp)
    else:
        logger.info("    No active data detected.")
        active_data = active_vert = None
    # Complete dicts :
    kw_brain_obj['vertices'], kw_brain_obj['faces'] = vertices, faces
    kw_brain_obj['lr_index'], kw_brain_obj['hemisphere'] = lr_index, hemisphere
    return _plt_src(sbj, kw_brain_obj, active_data, active_vert, sources,
                    kw_source_obj, kw_activation, show)


def mne_plot_source_space(fif_file, active_data=None, hemisphere='both',
                          kw_brain_obj={}, kw_source_obj={}, kw_activation={},
                          show=True):
    """Plot source space.

    Parameters
    ----------
    fif_file : string
        Full path to the .fif file.
    active_data : array_like | None
        The data to set to vertices.
    hemisphere : {'left', 'both', 'right'}
        The hemisphere to plot.
    kw_brain_obj : dict | {}
        Additional inputs to pass to the `BrainObj` class.
    kw_source_obj : dict | {}
        Additional inputs to pass to the `SourceObj` class.
    kw_activation : dict | {}
        Additional inputs to pass to the `BrainObj.add_activation` method.
    show : bool | False
        If True, the window of the `Brain` module is automatically displayed.
        If False, a BrainObj and a SourceObj are returned. Finally, if 'scene'
        a SceneObj is returned.

    Returns
    -------
    b_obj : BrainObj
        A predefined `BrainObj` (if `show=False`)
    s_obj : SourceObj
        A predefined `SourceObj`, hide by default (if `show=False`)
    """
    # Test that mne is installed and import :
    is_mne_installed(raise_error=True)
    from mne import read_source_spaces
    assert os.path.isfile(fif_file)
    src = read_source_spaces(fif_file)
    # Build vertices / faces :
    (vertices, faces, lr_index, active_vert,
     sources) = _extract_arrays_from_src(src, hemisphere, fact=1000.)
    # Complete dicts :
    kw_brain_obj['vertices'], kw_brain_obj['faces'] = vertices, faces
    kw_brain_obj['lr_index'], kw_brain_obj['hemisphere'] = lr_index, hemisphere
    file = os.path.splitext(os.path.split(fif_file)[1])[0]
    return _plt_src(file, kw_brain_obj, active_data, active_vert, sources,
                    kw_source_obj, kw_activation, show)



"""This script contains some usefull signal processing functions."""
import logging

import numpy as np
from scipy.signal import fftconvolve

from vispy.visuals.transforms import STTransform, NullTransform


__all__ = ('normalize', 'derivative', 'tkeo', 'zerocrossing', 'power_of_ten',
           'averaging', 'normalization', 'smoothing', 'smooth_3d')

logger = logging.getLogger('visbrain')


def normalize(x, tomin=0., tomax=1.):
    """Normalize the array x between tomin and tomax.

    Parameters
    ----------
    x : array_like
        The array to normalize
    tomin : int/float | 0.
        Minimum of returned array

    tomax : int/float | 1.
        Maximum of returned array

    Returns
    -------
    xn : array_like
        The normalized array
    """
    if x.size:
        x = np.float32(x)
        xm, xh = np.float32(x.min()), np.float32(x.max())
        if xm != xh:
            coef = (tomax - tomin) / (xh - xm)
            np.subtract(x, xh, out=x)
            np.multiply(x, coef, out=x)
            np.add(x, tomax, out=x)
            return x
            # return tomax - (((tomax - tomin) * (xh - x)) / (xh-xm))
        else:
            logger.debug("Normalization has been ignored because minimum and "
                         "maximum are both equal to " + str(xm))
            np.multiply(x, tomax, out=x)
            np.divide(x, xh, out=x)
            return x
    else:
        return x


def derivative(x, window, sf):
    """Compute first derivative of signal.

    Equivalent to np.gradient function

    Parameters
    ----------
    x : array_like
        Signal
    window : int
        Time (ms) window to compute first derivative
    sf : int
        Downsampling frequency
    """
    length = x.size
    step = int(window / (1000 / sf))
    tail = np.zeros(shape=(int(step / 2),))
    deriv = np.r_[tail, x[step:length] - x[0:length - step], tail]
    deriv = np.abs(deriv)
    # Check size
    if deriv.size < length:
        missing_pts = length - deriv.size
        tail = np.zeros(missing_pts)
        deriv = np.r_[deriv, tail]

    return deriv


def tkeo(x):
    """Calculate the TKEO of a given recording by using 2 samples.

    github.com/lvanderlinden/OnsetDetective/blob/master/OnsetDetective/tkeo.py

    Parameters
    ----------
    x : array_like
        Row vector of data.

    Returns
    -------
    a_tkeo : array_like
        Row vector containing the tkeo per sample.
    """
    # Create two temporary arrays of equal length, shifted 1 sample to the
    # right and left and squared:
    i = x[1:-1] * x[1:-1]
    j = x[2:] * x[:-2]

    # Calculate the difference between the two temporary arrays:
    a_tkeo = i - j
    return a_tkeo


def zerocrossing(data):
    """Find zero-crossings index of a signal.

    Parameters
    ----------
    x: array_like
        Data

    Returns
    -------
    index : array_like
        Row vector containing zero-crossing index.
    """
    pos = data > 0
    npos = ~pos
    return ((pos[:-1] & npos[1:]) | (npos[:-1] & pos[1:])).nonzero()[0] + 1


def power_of_ten(x, e=3):
    """Power of ten format.

    Parameters
    ----------
    x : float
        The floating point to transform.
    e : int | 2
        If x is over 10 ** -e and bellow 10 ** e, this function doesn't
        change the format.

    Returns
    -------
    xtronc: float
        The troncate version of x.
    power: int
        The power of ten to retrieve x.
    """
    sign = np.sign(x)
    x = np.abs(x)
    stx = str(x)
    if 0 < x <= 10 ** -e:  # x is a power of e- :
        if stx.find('e-') + 1:  # Format : 'xe-y'
            sp = stx.split('e-')
            return float(sp[0]), -int(sp[1])
        else:  # Format : 0.000x
            sp = stx.split('.')[1]
            id_l = 0
            while sp[id_l] == '0':
                id_l += 1
            id_l += 1
            return (sign * x) * (10 ** id_l), -id_l
    elif x >= 10 ** e:  # x is a power of e :
        if stx.find('e') + 1:  # Format : 'xey'
            sp = stx.split('e')
            return float(sp[0]), -int(sp[1])
        else:
            k = e
            while x % (10 ** k) != x:
                k += 1
            return (sign * x) / (10 ** (k - 1)), k - 1
    else:
        return sign * x, 0


def averaging(ts, n_window, axis=-1, overlap=0., window='flat'):
    """Take the mean of an ndarray.

    Parameters
    ----------
    ts : array_like
        Array of data to take the mean.
    n_window : int
        Number of sample per window.
    axis : int | -1
        Axis along which take the mean. By default, the last axis.
    overlap : float | None
        Overlap of successive window (0 <= overlap < 1). By default, no overlap
        is performed.
    window : {'flat', 'hanning', 'hamming', 'bartlett', 'blackman'}
        Windowing method.

    Returns
    -------
    average : array_like
        The averaged signal.
    """
    # Checking :
    assert isinstance(ts, np.ndarray)
    assert isinstance(axis, int) and axis <= ts.ndim - 1
    assert isinstance(n_window, int) and n_window < ts.shape[axis]
    assert isinstance(overlap, (float, int)) and 0. <= overlap < 1.
    assert window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']

    # Get axis :
    npts = ts.shape[axis]
    axis = ts.ndim - 1 if axis == -1 else axis

    # Get overlap step in samples :
    n_overlap = int(np.round(n_window * (1. - overlap)))

    # Build the index vector :
    ind = np.c_[np.arange(0, npts - n_window, n_overlap),
                np.arange(n_window, npts, n_overlap)]
    ind = np.vstack((ind, [npts - 1 - n_window, npts - 1]))  # add last window
    n_ind = ind.shape[0]

    # Get the window :
    if window == 'flat':  # moving average
        win = np.ones(n_window, 'd')
    else:
        win = eval('np.' + window + '(n_window)')

    rsh = tuple(1 if i != axis else -1 for i in range(ts.ndim))
    win = win.reshape(*rsh)

    # Define the averaging array :
    av_shape = tuple(k if i != axis else n_ind for i, k in enumerate(ts.shape))
    average = np.zeros(av_shape, dtype=float)

    # Compute averaging :
    sl_ts = [slice(None)] * ts.ndim
    sl_av = sl_ts.copy()
    for k in range(n_ind):
        sl_ts[axis], sl_av[axis] = slice(ind[k, 0], ind[k, 1]), slice(k, k + 1)
        average[tuple(sl_av)] += (ts[tuple(sl_ts)] * win).mean(axis=axis,
                                                               keepdims=True)

    return average


def normalization(data, axis=-1, norm=None, baseline=None):
    """Data normalization.

    Parameters
    ----------
    data : array_like
        Array of data.
    axis : int | -1
        Array along which to perform the normalization.
    norm : int | None
        The normalization type. Use :
            * 0 : no normalization
            * 1 : subtract the mean
            * 2 : divide by the mean
            * 3 : subtract then divide by the mean
            * 4 : subtract the mean then divide by deviation
    baseline : tuple | None
        Baseline period to consider. If None, the entire signal is used.

    Returns
    -------
    data_n : array_like
        The normalized array.
    """
    assert isinstance(data, np.ndarray)
    # assert norm in [None, ]

    # Take data in baseline (if defined) :
    if (baseline is not None) and (len(baseline) == 2):
        sl = [slice(None)] * data.ndim
        sl[axis] = slice(baseline[0], baseline[1])
        _data = data[tuple(sl)]
    else:
        _data = None

    if norm in [0, None]:  # don't normalize
        return data
    elif norm in [1, 2, 3, 4]:
        kw = {'axis': axis, 'keepdims': True}
        d_m = _data.mean(**kw) if _data is not None else data.mean(**kw)
        if norm == 1:  # subtract the mean
            data -= d_m
        elif norm == 2:  # divide by the mean
            d_m[d_m == 0] = 1.
            data /= d_m
        elif norm == 3:  # subtract then divide by the mean
            data -= d_m
            d_m[d_m == 0] = 1.
            data /= d_m
        elif norm == 4:  # z-score
            d_std = _data.mean(**kw) if _data is not None else data.mean(**kw)
            d_std[d_std == 0] = 1.
            data -= d_m
            data /= d_std


def smoothing(x, n_window=10, window='hanning'):
    """Smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    Parameters
    ----------
    x : array_like
        1-D array to smooth.
    n_window : int | 10
        Window length.
    window : string, array_like | 'hanning'
        Use either 'flat', 'hanning', 'hamming', 'bartlett', 'blackman' or pass
        a numpy array of length n_window.

    Returns
    -------
        The smoothed signal
    """
    n_window = int(n_window)
    assert isinstance(x, np.ndarray) and x.ndim == 1
    assert len(x) > n_window
    assert isinstance(window, (str, np.ndarray))
    if isinstance(window, str):
        assert window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']
    elif isinstance(window, np.ndarray):
        assert len(window) == n_window

    if n_window < 3:
        return x

    s = np.r_[2 * x[0] - x[n_window:1:-1], x, 2 * x[-1] - x[-1:-n_window:-1]]
    if window == 'flat':  # Moving average
        w = np.ones((n_window,))
    else:
        w = eval('np.' + window + '(n_window)')

    y = np.convolve(w / w.sum(), s, mode='same')
    return y[n_window - 1:-n_window + 1]


def smooth_3d(vol, smooth_factor=3, correct=True):
    """Smooth a 3-D volume.

    Parameters
    ----------
    vol : array_like
        The volume of shape (N, M, P)
    smooth_factor : int | 3
        The smoothing factor.

    Returns
    -------
    vol_smooth : array_like
        The smooth volume with the same shape as vol.
    """
    tf = NullTransform()
    # No smoothing :
    if (not isinstance(smooth_factor, int)) or (smooth_factor < 3):
        return vol, tf
    # Smoothing array :
    sz = np.full((3,), smooth_factor, dtype=int)
    smooth = np.ones([smooth_factor] * 3) / np.prod(sz)
    # Apply smoothing :
    sm = fftconvolve(vol, smooth, mode='same')
    if correct:
        # Get the shape of the vol and the one with 'full' convolution :
        vx, vy, vz = vol.shape
        vcx, vcy, vcz = np.array([vx, vy, vz]) + smooth_factor - 1
        # Define transform :
        sc = [vx / vcx, vy / vcy, vz / vcz]
        tr = .5 * np.array([smooth_factor] * 3)
        tf = STTransform(scale=sc, translate=tr)
    return sm, tf



"""Visual colorbar object using VisPy."""
import numpy as np
from vispy.scene import Node, visuals
from vispy import scene
import vispy.visuals.transforms as vist

from ...utils import array2colormap, color2tuple, FixedCam
from .CbarBase import CbarBase


class CbarVisual(CbarBase):
    """Create a colorbar using Vispy.

    Parameters
    ----------
    cmap : string | None
        Matplotlib colormap (like 'viridis', 'inferno'...).
    clim : tuple/list | None
        Colorbar limit. Every values under / over clim will
        clip.
    isvmin : bool | False
        Activate/deactivate vmin.
    vmin : float | None
        Every values under vmin will have the color defined
        using the under parameter.
    isvmax : bool | False
        Activate/deactivate vmax.
    vmax : float | None
        Every values over vmin will have the color defined
        using the over parameter.
    under : tuple/string | None
        Matplotlib color under vmin.
    over : tuple/string | None
        Matplotlib color over vmax.
    cblabel : string | ''
        Colorbar label.
    cbtxtsz : float | 5..
        Text size of the colorbar label.
    cbtxtsh : float | 2.3
        Shift for the colorbar label.
    txtcolor : string | 'white'
        Text color.
    txtsz : float | 3.
        Text size for clim/vmin/vmax text.
    txtsh : float | 1.2
        Shift for clim/vmin/vmax text.
    border : bool | True
        Display colorbar borders.
    bw : float | 2.
        Border width.
    limtxt : bool | True
        Display vmin/vmax text.
    bgcolor : tuple/string | (0., 0., 0.)
        Background color of the colorbar canvas.
    ndigits : int | 2
        Number of digits for the text.
    parent : VisPy | None
        VisPy parent to use.
    """

    def __init__(self, parent=None, **kwargs):
        """Init."""
        # _____________________ INIT _____________________
        self._n = 1000
        self._ratio = 4 / 5
        CbarBase.__init__(self, **kwargs)

        # _____________________ CANVAS _____________________
        if parent is None:
            # Define a canvas :
            self._canvas = scene.SceneCanvas(keys='interactive', show=False,
                                             resizable=True, dpi=600,
                                             bgcolor=self._bgcolor,
                                             size=(300, 900))
            self._wc = self._canvas.central_widget.add_view()
            parent = self._wc.scene
            # Define the camera :
            self._camera = FixedCam(rect=(-1.2, -1.2, 2.4, 2.4))
            self._wc.camera = self._camera
        self.parent = parent

        # _____________________ OBJECTS _____________________
        # --------------------- Node ---------------------
        # Define node parent and limit node :
        self._cbNode = Node(name='Colorbar', parent=parent)
        self._limNode = Node(name='Limits', parent=self._cbNode)
        # Rescale between (-1., 1.) :
        self._rsc = vist.STTransform(scale=(self._width, 2 / self._n, 1),
                                     translate=(0, -1., 0))
        # Set transformation to the node :
        self._cbNode.transform = self._rsc

        # --------------------- Image ---------------------
        self._mImage = visuals.Image(parent=self._cbNode, name='image')

        # --------------------- Border ---------------------
        pos = np.array([[0., 0., -3.],
                        [1., 0., -3.],
                        [1., 0., -3.],
                        [1., self._n, -3.],
                        [1., self._n, -3.],
                        [0., self._n, -3.],
                        [0., self._n, -3.],
                        [0., 0., -3.]])
        self._mBorder = visuals.Line(parent=self._cbNode, name='Border')
        self._mBorder.set_data(pos=pos, width=2., connect='segments',
                               color=self._txtcolor)
        self._mBorder.visible = self._border

        # --------------------- Labels ---------------------
        # Clim labels :
        self._mClimM = visuals.Text(parent=self._limNode, color=self._txtcolor,
                                    font_size=self._txtsz, name='Clim_M',
                                    anchor_x='left')
        self._mClimm = visuals.Text(parent=self._limNode, color=self._txtcolor,
                                    font_size=self._txtsz, name='Clim_m',
                                    anchor_x='left')

        # Cblabel :
        self._mcblabel = visuals.Text(parent=self._limNode, name='Cblabel',
                                      color=self._txtcolor, anchor_x='center',
                                      font_size=self._cbtxtsz)
        self._mcblabel.rotation = -90

        # Vmin/Vmax :
        self._vmMNode = Node(name='VminVmax', parent=self._limNode)
        self._mVm = visuals.Text(parent=self._vmMNode, color=self._txtcolor,
                                 font_size=self._ratio * self._txtsz,
                                 name='Vmin', anchor_x='left')
        self._mVM = visuals.Text(parent=self._vmMNode, color=self._txtcolor,
                                 font_size=self._ratio * self._txtsz,
                                 name='Vmax', anchor_x='left')

        # Build colorbar :
        self._build(True, 'all')

    # -------------------------------------------------------------------------
    #                             DEEP METHODS
    # -------------------------------------------------------------------------
    def _check(self):
        """Check variables."""
        try:
            # Check if clim is int/float :
            if self._clim is None:
                if not hasattr(self, '_minmax'):
                    self._minmax = (0., 1.)
                self._clim = self._minmax
            assert all([isinstance(k, (int, float)) for k in self._clim])
            # Check vmin/vmax :
            assert (self._vmin is None) or isinstance(self._vmin, (int, float))
            assert (self._vmax is None) or isinstance(self._vmax, (int, float))
            return True
        except:
            raise ValueError("Error in checking")

    def _setter(self, toset):
        # _____________________ IMAGE _____________________
        self._mImage.set_data(self._colormap)

        # _____________________ CLIM _____________________
        clim = np.array(self._clim)
        if toset in ['all', 'clim']:
            self._mClimm.text = str(self._digits(clim[0]))
            self._mClimM.text = str(self._digits(clim[1]))

        # _____________________ VMIN/VMAX _____________________
        # ------------ VMIN ------------
        if toset in ['all', 'vmin']:
            isnn = (self._vmin is not None) and self._isvmin
            if isnn and (clim[0] < self._vmin < clim[1]):
                self._mVm.visible = True
                self._mVm.text = str(self._digits(self._vmin))
                self._mVm.pos = (self.txtsh, self._conv(self._vmin), -3.)
            else:
                self._mVm.visible = False
        # ------------ VMAX ------------
        if toset in ['all', 'vmax']:
            isnn = (self._vmax is not None) and self._isvmax
            if isnn and (clim[0] < self._vmax < clim[1]):
                self._mVM.visible = True
                self._mVM.text = str(self._digits(self._vmax))
                self._mVM.pos = (self.txtsh, self._conv(self._vmax), -3.)
            else:
                self._mVM.visible = False

        # _____________________ TEXT _____________________
        # Text color :
        if toset in ['all', 'txtcolor']:
            self._mClimm.color = self._txtcolor
            self._mClimM.color = self._txtcolor
            self._mcblabel.color = self._txtcolor
            self._mVm.color = self._txtcolor
            self._mVM.color = self._txtcolor
            self._mBorder.set_data(color=self._txtcolor)
            self._mcblabel.color = self._txtcolor

        # Label :
        self._cblabel = '' if self._cblabel is None else self._cblabel
        if toset in ['all', 'cblabel']:
            self._mcblabel.text = self._cblabel
        if toset in ['all', 'cbtxtsh']:
            self._mcblabel.pos = (self.cbtxtsh, self._n / 2., -3.)
        if toset in ['all', 'cbtxtsz']:
            self._mcblabel.font_size = self._cbtxtsz
        # Clim/Vmin/Vmax font size and position :
        if toset in ['all', 'txtsz']:
            self._mClimm.font_size = self._txtsz
            self._mClimM.font_size = self._txtsz
            self._mVm.font_size = self._ratio * self._txtsz
            self._mVM.font_size = self._ratio * self._txtsz
        if toset in ['all', 'txtsh']:
            self._mClimM.pos = (self.txtsh, self._n, -3.)
            self._mClimm.pos = (self.txtsh, 0., -3.)
            if self._vmin is not None:
                self._mVm.pos = (self.txtsh, self._conv(self._vmin), 3.)
            if self._vmax is not None:
                self._mVM.pos = (self.txtsh, self._conv(self._vmax), 3.)

    def _build(self, needupdate=True, toset='all'):
        """Build the colormap."""
        if self._check() and needupdate:
            # Generate a sample vector between clim :
            sample = np.linspace(self._clim[0], self._clim[1], self._n,
                                 endpoint=True)
            # Avoid setting vmin/vmax if checkboxes inactives :
            vmin = self._vmin if self._isvmin else None
            vmax = self._vmax if self._isvmax else None
            # Get the associated colormap :
            cp = array2colormap(sample, cmap=self._cmap, vmin=vmin,
                                vmax=vmax, under=self._under, over=self._over)
            self._colormap = cp[:, np.newaxis, :]
        self._setter(toset)
        self._update()

    def _update(self):
        """Update VisPy objects."""
        self._mImage.update()
        self._mcblabel.update()
        self._mClimm.update()
        self._mClimM.update()
        self._mVm.update()
        self._mVM.update()
        self._mBorder.update()

    def _conv(self, value):
        clim = self._clim
        if clim[1] > clim[0]:
            return self._n * (value - clim[0]) / (clim[1] - clim[0])

    def _digits(self, value):
        if isinstance(self._ndigits, int):
            txt = np.round(value * 10**self._ndigits) / 10**self._ndigits
        else:
            txt = value
        return txt

    ###########################################################################
    ###########################################################################
    #                                 SETTINGS
    ###########################################################################
    ###########################################################################
    # ----------- NAME -----------
    @property
    def name(self):
        """Get the name value."""
        return self._name

    @name.setter
    def name(self, value):
        """Set name value."""
        self._name = value

    # ----------- BGCOLOR -----------
    @property
    def bgcolor(self):
        """Get the bgcolor value."""
        return self._bgcolor

    @bgcolor.setter
    def bgcolor(self, value):
        """Set bgcolor value."""
        self._bgcolor = value
        self._canvas.bgcolor = value
        self._canvas.update()

    # ----------- NDIGITS -----------
    @property
    def ndigits(self):
        """Get the ndigits value."""
        return self._ndigits

    @ndigits.setter
    def ndigits(self, value):
        """Set ndigits value."""
        bck = self._ndigits
        self._ndigits = value
        self._build(bck is not value, 'all')

    # ----------- WIDTH -----------
    @property
    def width(self):
        """Get the width value."""
        return self._width

    @width.setter
    def width(self, value):
        """Set width value."""
        self._width = value
        sc = self._cbNode.transform.scale
        tr = self._cbNode.transform.translate
        self._cbNode.transform.scale = (value, sc[1], sc[2], sc[3])
        self._cbNode.transform.translate = (-value, tr[1], tr[2], tr[3])

    # ----------- BORDER -----------
    @property
    def border(self):
        """Get the border value."""
        return self._border

    @border.setter
    def border(self, value):
        """Set border value."""
        self._border = value
        self._mBorder.visible = value

    # ----------- BW -----------
    @property
    def bw(self):
        """Get the bw value."""
        return self._bw

    @bw.setter
    def bw(self, value):
        """Set bw value."""
        self._bw = value
        self._mBorder.set_data(width=value)

    # ----------- LIMTXT -----------
    @property
    def limtxt(self):
        """Get the limtxt value."""
        return self._limtxt

    @limtxt.setter
    def limtxt(self, value):
        """Set limtxt value."""
        self._limtxt = value
        self._vmMNode.visible = value

    ###########################################################################
    ###########################################################################
    #                       CMAP/VMIN/VMAX/UNDER/OVER
    ###########################################################################
    ###########################################################################
    # ----------- CMAP -----------
    @property
    def cmap(self):
        """Get the cmap value."""
        return self._cmap

    @cmap.setter
    def cmap(self, value):
        """Set cmap value."""
        bck = self._cmap
        self._cmap = value
        self._build(bck is not value)

    # ----------- CLIM -----------
    @property
    def clim(self):
        """Get the clim value."""
        return self._clim

    @clim.setter
    def clim(self, value):
        """Set clim value."""
        bck = self._clim
        self._clim = value
        self._build(bck is not value, 'all')

    # ----------- ISVMIN -----------
    @property
    def isvmin(self):
        """Get the isvmin value."""
        return self._isvmin

    @isvmin.setter
    def isvmin(self, value):
        """Set isvmin value."""
        bck = self._isvmin
        self._isvmin = value
        self._build(bck is not value, 'vmin')

    # ----------- VMIN -----------
    @property
    def vmin(self):
        """Get the vmin value."""
        return self._vmin

    @vmin.setter
    def vmin(self, value):
        """Set vmin value."""
        bck = self._vmin
        self._vmin = value
        self._build(bck is not value, 'vmin')

    # ----------- UNDER -----------
    @property
    def under(self):
        """Get the under value."""
        return self._under

    @under.setter
    def under(self, value):
        """Set under value."""
        bck = self._under
        self._under = color2tuple(value, float)
        self._build(bck is not value)

    # ----------- ISVMAX -----------
    @property
    def isvmax(self):
        """Get the isvmax value."""
        return self._isvmax

    @isvmax.setter
    def isvmax(self, value):
        """Set isvmax value."""
        bck = self._isvmax
        self._isvmax = value
        self._build(bck is not value, 'vmax')

    # ----------- VMAX -----------
    @property
    def vmax(self):
        """Get the vmax value."""
        return self._vmax

    @vmax.setter
    def vmax(self, value):
        """Set vmax value."""
        bck = self._vmax
        self._vmax = value
        self._build(bck is not value, 'vmax')

    # ----------- OVER -----------
    @property
    def over(self):
        """Get the over value."""
        return self._over

    @over.setter
    def over(self, value):
        """Set over value."""
        bck = self._over
        self._over = color2tuple(value, float)
        self._build(bck is not value)

    ###########################################################################
    ###########################################################################
    #                                 TEXT
    ###########################################################################
    ###########################################################################
    # ----------- CBLABEL -----------
    @property
    def cblabel(self):
        """Get the cblabel value."""
        return self._cblabel

    @cblabel.setter
    def cblabel(self, value):
        """Set cblabel value."""
        self._cblabel = value
        self._build(False, 'cblabel')

    # ----------- CBTXTSZ -----------
    @property
    def cbtxtsz(self):
        """Get the cbtxtsz value."""
        return self._cbtxtsz

    @cbtxtsz.setter
    def cbtxtsz(self, value):
        """Set cbtxtsz value."""
        self._cbtxtsz = value
        self._build(False, 'cbtxtsz')

    # ----------- CBTXTSH -----------
    @property
    def cbtxtsh(self):
        """Get the cbtxtsh value."""
        return self._cbtxtsh

    @cbtxtsh.setter
    def cbtxtsh(self, value):
        """Set cbtxtsh value."""
        self._cbtxtsh = value
        self._build(False, 'cbtxtsh')

    # ----------- TXTCOLOR -----------
    @property
    def txtcolor(self):
        """Get the txtcolor value."""
        return self._txtcolor

    @txtcolor.setter
    def txtcolor(self, value):
        """Set txtcolor value."""
        self._txtcolor = color2tuple(value, float)
        self._build(False, 'txtcolor')

    # ----------- TXTSZ -----------
    @property
    def txtsz(self):
        """Get the txtsz value."""
        return self._txtsz

    @txtsz.setter
    def txtsz(self, value):
        """Set txtsz value."""
        self._txtsz = value
        self._build(False, 'txtsz')

    # ----------- TXTSH -----------
    @property
    def txtsh(self):
        """Get the txtsh value."""
        return self._txtsh

    @txtsh.setter
    def txtsh(self, value):
        """Set txtsh value."""
        self._txtsh = value
        self._build(False, 'txtsh')



"""Main class for settings managment."""
import numpy as np
from PyQt5 import QtWidgets

from sleep_scoring.utils import transient


class UiScoring(object):
    """Enable scoring using the table."""

    def __init__(self):
        """Init."""
        # Add / remove line :
        self._scoreAdd.clicked.connect(self._fcn_add_score_row)
        self._scoreRm.clicked.connect(self._fcn_rm_score_row)

        # Table edited :
        self._scoreTable.cellChanged.connect(self._fcn_score_to_hypno)

    ##########################################################################
    # UPDATE SCORE <=> HYPNO
    ##########################################################################
    def _fcn_hypno_to_score(self):
        """Update hypno table from hypno data."""
        self._hypno = self._hyp.gui_to_hyp()
        # Avoid updating data while setting cell :
        self._scoreSet = False
        items = ['Wake', 'N1', 'N2', 'N3', 'REM', 'Art']
        # Remove every info in the table :
        self._scoreTable.setRowCount(0)
        # Find unit conversion :
        fact = self._get_fact_from_unit()
        # Find transients :
        _, idx, stages = transient(self._hypno, self._time / fact)
        idx = np.round(10. * idx) / 10.
        # Set length of the table :
        self._scoreTable.setRowCount(len(stages))
        # Fill the table :
        for k in range(len(stages)):
            # Add stage start / end :
            self._scoreTable.setItem(k, 0, QtWidgets.QTableWidgetItem(
                str(idx[k, 0])))
            self._scoreTable.setItem(k, 1, QtWidgets.QTableWidgetItem(
                str(idx[k, 1])))
            # Add stage :
            self._scoreTable.setItem(k, 2, QtWidgets.QTableWidgetItem(
                items[stages[k]]))
        self._scoreSet = True

    def _fcn_score_to_hypno(self):
        """Update hypno data from hypno score."""
        if self._scoreSet:
            # Reset hypnogram :
            self._hypno = np.zeros((len(self._time)), dtype=np.float32)
            # Loop over table row :
            for k in range(self._scoreTable.rowCount()):
                # Get tstart / tend / stage :
                tstart, tend, stage = self._get_score_marker(k)
                # Update pos if not None :
                if tstart is not None:
                    self._hypno[tstart:tend] = stage
                    self._hyp.set_stage(tstart, tend, stage)
            self._hyp.edit.update()
            # Update sleep info :
            self._fcn_info_update()

    def _get_score_marker(self, idx):
        """Get a specific row dat.

        This function insure that the edited line is complete and is properly
        formated.

        Parameters
        ----------
        idx : int
            The row from which get data.

        Returns
        -------
        tstart : float
            Time start (in sample)
        tend : float
            Time end (in sample).
        stage : int
            The stage.
        """
        it = {'art': -1., 'wake': 0., 'n1': 1., 'n2': 2., 'n3': 3., 'rem': 4.}
        # Get unit :
        fact = self._get_fact_from_unit()
        # Get selected row :
        # idx = self._scoreTable.currentRow()
        # ============= NON EMPTY ITEM =============
        # Define error message if bad editing :
        errmsg = "\nTable score error. Starting and ending time must be " + \
                 "float numbers (with time start < time end) and stage " + \
                 "must be Wake, N1, N2, N3, REM or Art"
        # Get row data and update if possible:
        tstart_item = self._scoreTable.item(idx, 0)
        tend_item = self._scoreTable.item(idx, 1)
        stage_item = self._scoreTable.item(idx, 2)
        if tstart_item and tend_item and stage_item:
            # ============= PROPER FORMAT =============
            if all([bool(str(tstart_item.text())), bool(str(tend_item.text())),
                    str(stage_item.text()).lower() in it.keys()]):
                try:
                    # Get start / end / stage :
                    tstart = int(float(str(tstart_item.text(
                    ))) * fact * self._sf)
                    tend = int(float(str(tend_item.text())) * fact * self._sf)
                    stage = it[str(stage_item.text()).lower()]

                    return tstart, tend, stage
                except:
                    raise ValueError(errmsg)
            else:
                raise ValueError(errmsg)
        else:
            return None, None, None

    ##########################################################################
    # EDITING TABLE
    ##########################################################################
    def _fcn_add_score_row(self):
        """Add a row to the table."""
        # Increase length :
        self._scoreTable.setRowCount(self._scoreTable.rowCount() + 1)

    def _fcn_rm_score_row(self):
        """Remove selected row."""
        # Remove row :
        self._scoreTable.removeRow(self._scoreTable.currentRow())
        # Update hypnogram from table :
        self._fcn_score_to_hypno()



# -*- coding: utf-8 -*-

"""Group functions for automatic detection of sleep parameters.

Perform:
- REM detection
- Muscle Twitches detection
- Spindles detection
- Slow wave detection
- KCs detection
- Peak detection
"""
import numpy as np
from scipy.signal import hilbert, detrend, welch

from ..filtering import filt, morlet, morlet_power
from ..sigproc import derivative, tkeo, smoothing, normalization
from .event import (_events_distance_fill, _index_to_events, _events_to_index)

__all__ = ('kcdetect', 'spindlesdetect', 'remdetect', 'slowwavedetect',
           'mtdetect', 'peakdetect')

###########################################################################
# K-COMPLEX DETECTION
###########################################################################


def kcdetect(data, sf, proba_thr, amp_thr, hypno, nrem_only, tmin, tmax,
             kc_min_amp, kc_max_amp, fmin=.5, fmax=4., delta_thr=.75,
             smoothing_s=20, spindles_thresh=2., range_spin_sec=20,
             min_distance_ms=500.):
    """Perform a K-complex detection.

    Parameters
    ----------
    data : array_like
        eeg signal (preferably central electrodes)
    sf : float
        Downsampling frequency
    proba_thr : float
        Probability threshold (between 0 and 1)
    amp_thr : float
        Amplitude threshold
    hypno : array_like
        Hypnogram vector, same length as data
        Vector with only 0 if no hypnogram is loaded
    nrem_only : bool
        Perfom detection only on NREM sleep period
    tmin : float
        Minimum duration (ms) of K-complex
    tmax : float
        Maximum duration (ms) of K-complex
    kc_min_amp : float
        Minimum amplitude of K-complex
    kc_max_amp : float
        Maximum amplitude of K-complex
    fmin : float | .5
        High-pass cutoff frequency
    fmax : float | 4.
        Low-pass cutoff frequency
    delta_thr : float | .75
        Delta normalized power threshold. Value must be between 0 and 1.
        0 = No thresholding by delta bandpower
    smoothing_s : int | 20
        Time window (sec) for smoothing of delta band power
    spindles_thresh : float | 2.
        Number of standard deviations to compute spindles detection
    range_spin_sec : int | 20
        Duration of lookahead window for spindles detection (sec)
        Check for spindles that are comprised within
        -range_spin_sec/2 < KC < range_spin_sec/2
    min_distance_ms : float | 500.
        Minimum distance (ms) between two unique K-complexes

    Returns
    -------
    idx_kc : array_like
        Indices of detected K-complexes of shape (n_events, 2)
    """
    # Find if hypnogram is loaded :
    hyploaded = True if np.unique(hypno).size > 1 and nrem_only else False

    # PRE DETECTION
    # Compute delta band power using wavelet
    freqs = np.array([0.1, 4., 8., 12., 16., 30.])
    delta_npow = morlet_power(data, freqs, sf, norm=True)[0]
    delta_nfpow = smoothing(delta_npow, smoothing_s * sf)
    idx_no_delta = np.where(delta_nfpow < delta_thr)[0]
    idx_loc_delta = np.where(delta_npow > np.median(delta_npow))[0]

    # MAIN DETECTION
    # Bandpass filtering
    sig_filt = filt(sf, np.array([fmin, fmax]), data)
    # Taiger-Keaser energy operator
    sig_tkeo = tkeo(sig_filt)
    # Define hard and soft thresholds
    hard_thr = np.nanmean(sig_tkeo) + amp_thr * np.nanstd(sig_tkeo)
    soft_thr = 0.8 * hard_thr

    with np.errstate(divide='ignore', invalid='ignore'):
        idx_hard = np.where(sig_tkeo > hard_thr)[0]
        idx_soft = np.where(sig_tkeo > soft_thr)[0]

    # Find threshold-crossing indices of soft threshold
    idx_zc_soft = _events_to_index(idx_soft).flatten()

    if idx_hard.size == 0:
        return np.array([], dtype=int)

    # Initialize K-complexes index vector
    idx_kc = np.array([], dtype=int)
    # Fill gap between events separated by less than min_distance_ms
    idx_hard = _events_distance_fill(idx_hard, min_distance_ms, sf)
    # Get where K-complex start / end :
    idx_start, idx_stop = _events_to_index(idx_hard).T

    # Find true beginning / end using soft threshold
    for s in idx_start:
        d = s - idx_zc_soft
        soft_beg = d[d > 0].min()
        soft_end = np.abs(d[d < 0]).min()
        idx_kc = np.append(idx_kc, np.arange(s - soft_beg, s + soft_end))

    # Check if spindles are present in range_spin_sec
    idx_spin = spindlesdetect(data, sf, spindles_thresh, hypno, False)[0]
    idx_start, idx_stop = _events_to_index(idx_kc).T
    spin_bool = np.array([], dtype=np.bool)

    for idx, val in enumerate(idx_start):
        step = 0.5 * range_spin_sec * sf
        is_spin = np.in1d(np.arange(val - step, val + step, 1),
                          idx_spin, assume_unique=True)
        spin_bool = np.append(spin_bool, any(is_spin))

    kc_spin = np.where(spin_bool)[0]
    idx_kc_spin = _index_to_events(np.c_[idx_start, idx_stop][kc_spin])

    # Compute probability
    proba = np.zeros(shape=data.shape)
    proba[idx_kc] += 0.1
    proba[idx_no_delta] += 0.1
    proba[idx_loc_delta] += 0.1
    proba[idx_kc_spin] += 0.1

    if hyploaded:
        proba[hypno == -1] += -0.1
        proba[hypno == 0] += -0.2
        proba[hypno == 1] += 0
        proba[hypno == 2] += 0.1
        proba[hypno == 3] += -0.1
        proba[hypno == 4] += -0.2

    # Smooth and normalize probability vector
    proba = proba / 0.5 if hyploaded else proba / 0.4
    proba = smoothing(proba, sf)
    # Keep only proba >= proba_thr (user defined threshold)
    idx_kc = np.intersect1d(idx_kc, np.where(proba >= proba_thr)[0], True)

    if idx_kc.size == 0:
        return np.array([], dtype=int)

    # Morphological criteria
    idx_start, idx_stop = _events_to_index(idx_kc).T
    duration_ms = (idx_stop - idx_start) * (1000 / sf)

    # Remove events with bad duration
    good_dur = np.where(np.logical_and(duration_ms > tmin,
                                       duration_ms < tmax))[0]
    idx_kc = _index_to_events(np.c_[idx_start, idx_stop][good_dur])

    # Remove events with bad amplitude
    idx_start, idx_stop = _events_to_index(idx_kc).T
    amp = np.zeros(shape=idx_start.size)
    for i, (start, stop) in enumerate(zip(idx_start, idx_stop)):
        amp[i] = np.ptp(data[start:stop])
    good_amp = np.where(np.logical_and(amp > kc_min_amp,
                                       amp < kc_max_amp))[0]

    return np.c_[idx_start, idx_stop][good_amp]


###########################################################################
# SPINDLES DETECTION
###########################################################################

def spindlesdetect(data, sf, threshold, hypno, nrem_only, fmin=12., fmax=14.,
                   tmin=300, tmax=3000, method='wavelet', min_distance_ms=300,
                   sigma_thr=0.2, adapt_band=True, return_full=False):
    """Perform a sleep spindles detection.

    Parameters
    ----------
    data : array_like
        eeg signal (preferably central electrodes)
    sf : float
        Downsampling frequency
    threshold : float
        Number of standard deviation to use as threshold
        Threshold is defined as: mean + X * std(derivative)
    hypno : array_like
        Hypnogram vector, same length as data
        Vector with only 0 if no hypnogram is loaded
    nrem_only : bool
        Perfom detection only on NREM sleep period
    fmin : float | 12
        Lower bandpass frequency
    fmax : float | 14
        Higher bandpass frequency
    method: {'wavelet', 'hilbert'}
        Method to extract complex decomposition. Use either 'hilbert' or
        'wavelet'.
    min_distance_ms : int | 300
        Minimum distance (in ms) between two spindles to consider them as
        two distinct spindles
    sigma_thr : float | 0.2
        Sigma band-wise normalized power threshold (between 0 and 1)
    adapt_band : bool | True
        If true, adapt sigma band limit by finding the peak sigma freq.
    return_full : bool | False
        If true, return more variables (start, stop, sigma, hard and soft
        thresh) Used in function write_fig_spindles

    Returns
    -------
    idx_spindles : array_like
        Indices of detected spindles of shape (n_events, 2)
    """
    # Pre-detection
    if adapt_band:
        # Find peak sigma frequency
        f, pxx_den = welch(data, sf)
        mfs = f[pxx_den == pxx_den[np.where((f >= 11) & (f < 16))].max()][0]
        fmin = mfs - 1
        fmax = mfs + 1

    # Compute relative sigma power
    freqs = np.array([0.5, 4., 8., fmin, fmax])
    sigma_npow = morlet_power(data, freqs, sf, norm=True)[-1]
    sigma_nfpow = smoothing(sigma_npow, sf * (tmin / 1000))
    # Vector of sigma power supra-threshold values
    idx_sigma = np.where(sigma_nfpow > sigma_thr)[0]

    # Get complex decomposition of filtered data :
    if method == 'hilbert':
        # Bandpass filter
        data_filt = filt(sf, [fmin, fmax], data, order=4)
        if data.size % 2:
            analytic = hilbert(data_filt)
        else:
            analytic = hilbert(data_filt[:-1], len(data_filt))
    elif method == 'wavelet':
        analytic = morlet(data, sf, np.mean([fmin, fmax]))

    # Get envelope
    amplitude = np.abs(analytic)

    # Check "Detect only for NREM sleep"
    if np.unique(hypno).size > 1 and nrem_only:
        idx_zero = np.where(np.logical_or(hypno < 1, hypno == 4))[0]
        amplitude[idx_zero] = np.nan
        length = max(data.shape) - idx_zero.size
    else:
        length = max(data.shape)

    # Define hard and soft thresholds
    hard_thr = np.nanmean(amplitude) + threshold * np.nanstd(amplitude)
    soft_thr = 0.5 * hard_thr

    with np.errstate(divide='ignore', invalid='ignore'):
        idx_hard = np.where(amplitude > hard_thr)[0]
        idx_soft = np.where(amplitude > soft_thr)[0]

    # Find threshold-crossing indices of soft threshold
    idx_zc_soft = _events_to_index(idx_soft).flatten()

    if idx_hard.size > 0:
        # Initialize spindles vector
        idx_spindles = np.array([], dtype=int)

        # Keep only period with high relative sigma power
        idx_hard = np.intersect1d(idx_hard, idx_sigma, True)

        # Fill gap between events separated by less than min_distance_ms
        idx_hard = _events_distance_fill(idx_hard, min_distance_ms, sf)

        # Get where spindles start / end :
        idx_start, idx_stop = _events_to_index(idx_hard).T

        # Find true beginning / end using soft threshold
        for s in idx_start:
            d = s - idx_zc_soft
            # Find distance to nearest soft threshold crossing before start
            soft_beg = d[d > 0].min()
            # Find distance to nearest soft threshold crossing after end
            soft_end = np.abs(d[d < 0]).min()
            idx_spindles = np.append(idx_spindles, np.arange(
                                     s - soft_beg, s + soft_end))

        # Fill gap between events separated by less than min_distance_ms
        idx_spindles = _events_distance_fill(idx_spindles, min_distance_ms, sf)

        # Get duration
        idx_start, idx_stop = _events_to_index(idx_spindles).T
        duration_ms = (idx_stop - idx_start) * (1000 / sf)

        # Remove events with bad duration
        good_dur = np.where(np.logical_and(duration_ms > tmin,
                                           duration_ms < tmax))[0]

        if idx_spindles.size == 0:
            return np.array([], dtype=int)

        if return_full:
            # Compute number, duration, density
            idx_start, idx_stop = np.c_[idx_start, idx_stop][good_dur].T
            number = idx_start.size
            duration_ms = (idx_stop - idx_start) * (1000 / sf)
            density = number / (length / sf / 60.)

            # Compute mean power of each spindles
            pwrs = np.zeros(shape=number)
            for i, (start, stop) in enumerate(zip(idx_start, idx_stop)):
                ind_pwr = morlet_power(data[start:stop], [fmin, fmax], sf,
                                       norm=False)[0]
                pwrs[i] = np.mean(ind_pwr)
            # Normalize by dividing by the mean
            normalization(pwrs, norm=2)

            return (idx_spindles, number, density, duration_ms, pwrs,
                    idx_start, idx_stop, hard_thr, soft_thr, idx_sigma,
                    fmin, fmax, sigma_nfpow, amplitude, sigma_thr)
        else:
            return np.c_[idx_start, idx_stop][good_dur]


###########################################################################
# REM DETECTION
###########################################################################


def remdetect(data, sf, hypno, rem_only, threshold, tmin=300, tmax=800,
              min_distance_ms=300, smoothing_ms=200, deriv_ms=50):
    """Perform a rapid eye movement (REM) detection.

    Function to perform a semi-automatic detection of rapid eye movements
    (REM) during REM sleep.

    Parameters
    ----------
    data: array_like
        EOG signal
    sf: float
        Downsampling frequency
    hypno: array_like
        Hypnogram vector, same length as data
        Vector with only 0 if no hypnogram is loaded
    rem_only: bool
        Perfom detection only on REM sleep period
    threshold: float
        Number of standard deviation of the derivative signal
        Threshold is defined as: mean + X * std(derivative)
    tmin : int | 300
        Minimum duration (ms) of rapid eye movement
    tmax : int | 1500
        Maximum duration (ms) of rapid eye movement
    min_distance_ms : int | 300
        Minimum distance (ms) between two saccades to consider them as two
        distinct events.
    smoothing_ms : int | 200 (= 5 Hz)
        Time (ms) window of the smoothing.
    deriv_ms : int | 50
        Time (ms) window of derivative computation

    Returns
    -------
    idx_rem: array_like
        Indices of detected REMs of shape (n_events, 2)
    """
    # Compute relative beta power
    freqs = np.array([0.5, 4., 8., 12, 40])
    beta_npow = morlet_power(data, freqs, sf, norm=True)[-1]
    beta_nfpow = smoothing(beta_npow, sf * (tmin / 1000))
    # Vector of beta power supra-threshold values
    idx_beta = np.where(beta_nfpow < np.percentile(beta_nfpow, 60))[0]

    # Compute smoothed derivative
    sm_sig = smoothing(data, sf * (smoothing_ms / 1000))
    deriv = derivative(sm_sig, deriv_ms, sf)
    deriv = smoothing(deriv, sf * (smoothing_ms / 1000))

    if rem_only and 4 in hypno:
        idx_zero = np.where(hypno < 4)[0]
        deriv[idx_zero] = np.nan

    # Define hard and soft thresholds
    hard_thr = np.nanmean(deriv) + threshold * np.nanstd(deriv)
    soft_thr = 0.5 * hard_thr

    with np.errstate(divide='ignore', invalid='ignore'):
        idx_hard = np.where(deriv > hard_thr)[0]
        idx_soft = np.where(deriv > soft_thr)[0]

    # Find threshold-crossing indices of soft threshold
    idx_zc_soft = _events_to_index(idx_soft).flatten()

    if idx_hard.size == 0:
        return np.array([], dtype=int)

    # Initialize rem vector
    idx_rem = np.array([], dtype=int)

    # Keep only period with low relative beta power (i.e. remove artefact)
    idx_hard = np.intersect1d(idx_hard, idx_beta, True)

    # Fill gap between events separated by less than min_distance_ms
    idx_hard = _events_distance_fill(idx_hard, min_distance_ms, sf)

    # Get where spindles start / end :
    idx_start, idx_stop = _events_to_index(idx_hard).T

    # Find true beginning / end using soft threshold
    for s in idx_start:
        d = s - idx_zc_soft
        # Find distance to nearest soft threshold crossing before start
        soft_beg = d[d > 0].min()
        # Find distance to nearest soft threshold crossing after end
        soft_end = np.abs(d[d < 0]).min()
        idx_rem = np.append(idx_rem, np.arange(s - soft_beg, s + soft_end))

    # Fill gap between events separated by less than min_distance_ms
    idx_rem = _events_distance_fill(idx_rem, min_distance_ms, sf)

    # Get duration
    idx_start, idx_stop = _events_to_index(idx_rem).T
    duration_ms = (idx_stop - idx_start) * (1000 / sf)

    # Remove events with bad duration
    good_dur = np.where(np.logical_and(duration_ms > tmin,
                                       duration_ms < tmax))[0]

    return np.c_[idx_start, idx_stop][good_dur]


###########################################################################
# SLOW WAVE DETECTION
###########################################################################


def slowwavedetect(data, sf, threshold, min_amp=70., max_amp=400., tmin=1000.,
                   fmin=.5, fmax=4., smoothing_s=20):
    """Perform a Slow Wave detection.

    Parameters
    ----------
    data : array_like
        eeg signal (preferably frontal electrodes)
    sf : float
        Downsampling frequency
    threshold : float
        First threshold: bandwise-normalized delta power Value must be between
        0 and 1.
    min_amp : float | 70.
        Secondary threshold: minimum amplitude (uV) of the raw signal.
        Slow waves are generally defined by amplitude > 70 uV.
    max_amp : float | 400.
        Maximum amplitude of slow wave
    tmin : float | 1000.
        Minimum duration (ms) of slow waves
    fmin  : float | .5
        High-pass frequency
    fmax  : float | 2.
        Low-pass frequency
    smoothing_s  : int | 20
        Smoothing window in seconds

    Returns
    -------
    idx_sw : array_like
        Indices of slow waves of shape (n_events, 2)
    """
    filt_fmax = np.minimum(45, sf / 2.0 - 0.75)  # protect Nyquist
    data_filt = filt(sf, [.1, filt_fmax], data)

    # Compute relative delta band-power
    delta_nfpow = morlet_power(data_filt, [fmin, fmax, 8, 12, 16, 30], sf,
                               norm=True)[0, :]
    delta_nfpow = smoothing(delta_nfpow, smoothing_s * sf)

    # Normalized power criteria
    idx_sw = np.where(delta_nfpow > threshold)[0]

    if idx_sw.size == 0:
        return np.array([], dtype=int)

    # Get where slow waves start / end :
    idx_start, idx_stop = _events_to_index(idx_sw).T
    duration_ms = (idx_stop - idx_start) * (1000 / sf)

    # Check amplitude and duration
    amp = np.zeros(shape=idx_start.shape)
    for idx, (start, stop) in enumerate(zip(idx_start, idx_stop)):
        amp[idx] = np.ptp(data[start:stop])
    good_amp = np.where(np.logical_and(amp > min_amp, amp < max_amp))[0]
    good_dur = np.where(duration_ms > tmin)[0]
    good_event = np.intersect1d(good_amp, good_dur, True)
    idx_sw = np.c_[idx_start, idx_stop][good_event]

    if idx_sw.size == 0:
        return np.array([], dtype=int)

    return idx_sw


###########################################################################
# MUSCLE TWITCHES DETECTION
###########################################################################


def mtdetect(data, sf, threshold, hypno, rem_only, fmin=0., fmax=50.,
             tmin=800, tmax=2500, min_distance_ms=1000, min_amp=50,
             max_amp=400):
    """Perform a detection of muscle twitches (MT).

    Sampling frequency must be at least 1000 Hz.

    Parameters
    ----------
    data : array_like
        EMG signal
    sf : float
        Downsampling frequency
    threshold : float
        Number of standard deviation to use as threshold
        Threshold is defined as: mean + X * std(hilbert envelope)
    hypno : array_like
        Hypnogram vector, same length as data
        Vector with only 0 if no hypnogram is loaded
    rem_only : bool
        Perfom detection only on NREM sleep period
    fmin : float | 0.
        Lower bandpass frequency
    fmax : float | 50.
        Higher bandpass frequency
    tmin : int | 800
        Minimum duration (ms) of MT
    tmax : int | 2500
        Maximum duration (ms) of MT
    min_distance_ms : int | 1000
        Minimum distance (in ms) between 2 MTs to consider them as
        two distinct events
    min_amp : int | 50
        Minimum amplitude of Muscle Twitches
    max_amp : int | 400
        Maximum amplitude of Muscle Twitches. Above this threshold,
        detected events are probably artefacts

    Returns
    -------
    idx_mt : array_like
        Indices of MTs of shape (n_events, 2)
    """
    # PRE DETECTION
    # Morlet envelope
    analytic = morlet(data, sf, np.mean([fmin, fmax]))
    amplitude = np.abs(analytic)
    amplitude = smoothing(amplitude, sf * (tmin / 1000))
    # Morlet power in delta band
    delta_nfpow = morlet_power(data, [0.5, 4], sf, norm=False)
    idx_high_delta = np.where(delta_nfpow > np.percentile(delta_nfpow, 75))[0]

    if rem_only and 4 in hypno:
        idx_zero = np.where(hypno < 4)[0]
        amplitude[idx_zero] = np.nan

    # Define hard threshold
    hard_thr = np.nanmean(amplitude) + threshold * np.nanstd(amplitude)

    with np.errstate(divide='ignore', invalid='ignore'):
        idx_hard = np.where(amplitude > hard_thr)[0]

    if idx_hard.size == 0:
        return np.array([], dtype=int)

    # Keep only MT in period with low relative delta power
    idx_hard = np.setdiff1d(idx_hard, idx_high_delta, True)

    # Fill gap between events separated by less than min_distance_ms
    idx_hard = _events_distance_fill(idx_hard, min_distance_ms, sf)

    # MORPHOLOGICAL CRITERIA
    idx_start, idx_stop = _events_to_index(idx_hard).T
    duration_ms = (idx_stop - idx_start) * (1000 / sf)

    # Remove events with bad duration
    good_dur = np.where(np.logical_and(duration_ms > tmin,
                                       duration_ms < tmax))[0]
    idx_mt = _index_to_events(np.c_[idx_start, idx_stop][good_dur])

    # Remove events with bad amplitude
    idx_start, idx_stop = _events_to_index(idx_mt).T
    amp = np.zeros(shape=idx_start.size)
    for i, (start, stop) in enumerate(zip(idx_start, idx_stop)):
        amp[i] = np.ptp(data[start:stop])
    good_amp = np.where(np.logical_and(amp > min_amp,
                                       amp < max_amp))[0]
    idx_mt = np.c_[idx_start, idx_stop][good_amp]

    # Compute number, duration, density
    if idx_mt.size == 0:
        return np.array([], dtype=int)

    return idx_mt


###########################################################################
# PEAKS DETECTION
###########################################################################


def peakdetect(sf, y_axis, x_axis=None, lookahead=200, delta=1., get='max',
               threshold='auto'):
    """Perform a peak detection.

    Converted from/based on a MATLAB script at:
    http://billauer.co.il/peakdet.html
    Original script :
    https://github.com/DiamondLightSource/auto_tomo_calibration-experimental/
    blob/master/old_code_scripts/peak_detect.py

    function for detecting local maxima and minima in a signal.
    Discovers peaks by searching for values which are surrounded by lower
    or larger values for maxima and minima respectively

    Parameters
    ----------
    sf : float
        The sampling frequency.
    y_axis : array_like
        Row vector containing the data.
    x_axis : array_like
        Row vector for the time axis. If omitted an index of the y_axis is
        used.
    lookahead : int | 200
        Distance to look ahead from a peak candidate to determine if
        it is the actual peak.
        '(samples / period) / f' where '4 >= f >= 1.25' might be a good
        value
    delta : float | 1.
        This specifies a minimum difference between a peak and the
        following points, before a peak may be considered a peak. Useful
        to hinder the function from picking up false peaks towards to end
        of the signal. To work well delta should be set to
        delta >= RMSnoise * 5.
        When omitted delta function causes a 20% decrease in speed.
        When used Correctly it can double the speed of the function
    get : string | 'max'
        Get either minimum values ('min'), maximum ('max') or min and max
        ('minmax').
    threshold : string/float | None
        Use a threshold to ignore values. Use None for no threshold, 'auto'
        to use the signal deviation or a float number for specific
        threshold.

    Returns
    -------
    index : array_like
        A vector containing peak indices of shape (n_events, 2)
    """
    # ============== CHECK DATA ==============
    if x_axis is None:
        x_axis = range(len(y_axis))
    # Check length :
    if len(y_axis) != len(x_axis):
        raise ValueError("Input vectors y_axis and x_axis must have same "
                         "length")
    # Needs to be a numpy array
    y_axis, x_axis = np.asarray(y_axis), np.asarray(x_axis)

    # store data length for later use
    length = len(y_axis)

    # Lookahead  & delta checking :
    if lookahead < 1:
        raise ValueError("Lookahead must be '1' or above in value")
    if not (np.isscalar(delta) and delta >= 0):
        raise ValueError("delta must be a positive number")

    # Check get :
    if get not in ['min', 'max', 'minmax']:
        raise ValueError("The get parameter must either be 'min', 'max' or"
                         " 'minmax'")

    # ============== PRE-ALLOCATION ==============
    max_peaks, min_peaks = [], []
    dump = []   # Used to pop the first hit which almost always is false

    # maxima and minima candidates are temporarily stored in
    # mx and mn respectively
    mn, mx = np.Inf, -np.Inf

    # ============== THRESHOLD ==============
    if threshold is not None:
        if threshold == 'auto':
            threshold = np.std(y_axis)
        # Detrend / demean y-axis :
        y_axisp = detrend(y_axis)
        y_axisp -= y_axisp.mean()
        # Find values above threshold :
        above = np.abs(y_axisp) >= threshold
        zp = zip(np.arange(length)[above], x_axis[above], y_axis[above])
    else:
        zp = zip(np.arange(length)[:-lookahead], x_axis[:-lookahead],
                 y_axis[:-lookahead])

    # ============== FIND MIN / MAX PEAKS ==============
    # Only detect peak if there is 'lookahead' amount of points after it
    for index, x, y in zp:
        if y > mx:
            mx = y
        if y < mn:
            mn = y

        # ==== Look for max ====
        if y < mx - delta and mx != np.Inf:
            # Maxima peak candidate found
            # look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index + lookahead].max() < mx:
                max_peaks.append(index)
                dump.append(True)
                # set algorithm to only find minima now
                mx = np.Inf
                mn = np.Inf
                if index + lookahead >= length:
                    # end is within lookahead no more peaks can be found
                    break
                continue

        # ==== Look for max ====
        if y > mn + delta and mn != -np.Inf:
            # Minima peak candidate found
            # look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index + lookahead].min() > mn:
                min_peaks.append(index)
                dump.append(False)
                # set algorithm to only find maxima now
                mn = -np.Inf
                mx = -np.Inf
                if index + lookahead >= length:
                    # end is within lookahead no more peaks can be found
                    break

    if min_peaks and max_peaks:
        # ============== CLEAN ==============
        # Remove the false hit on the first value of the y_axis
        if threshold is None:
            if dump[0]:
                max_peaks.pop(0)
            else:
                min_peaks.pop(0)
            del dump

        # ============== MIN / MAX / MINMAX ==============
        if get == 'max':
            index = np.array(max_peaks)
        elif get == 'min':
            index = np.array(min_peaks)
        elif get == 'minmax':
            index = np.vstack((min_peaks, max_peaks))

        return np.c_[index, index]
    else:
        return np.array([])
